import collections
import collections.abc
import types
import typing
from dataclasses import InitVar, dataclass
from datetime import datetime

import pytest
import typing_extensions

from mashumaro import DataClassDictMixin
from mashumaro.core.const import PY_310_MIN
from mashumaro.core.meta.code.builder import CodeBuilder

# noinspection PyProtectedMember
from mashumaro.core.meta.helpers import (
    collect_type_params,
    get_args,
    get_class_that_defines_field,
    get_class_that_defines_method,
    get_function_arg_annotation,
    get_function_return_annotation,
    get_generic_name,
    get_literal_values,
    get_type_annotations,
    get_type_origin,
    get_type_var_default,
    hash_type_args,
    is_annotated,
    is_dataclass_dict_mixin,
    is_dataclass_dict_mixin_subclass,
    is_dialect_subclass,
    is_generic,
    is_hashable,
    is_hashable_type,
    is_init_var,
    is_literal,
    is_named_tuple,
    is_new_type,
    is_optional,
    is_self,
    is_type_var_any,
    is_union,
    not_none_type_arg,
    resolve_type_params,
    substitute_type_params,
    type_name,
    type_var_has_default,
)
from mashumaro.core.meta.types.common import (
    FieldContext,
    ValueSpec,
    ensure_generic_collection,
    ensure_mapping_key_type_hashable,
)
from mashumaro.dialect import Dialect
from mashumaro.exceptions import UnserializableField
from mashumaro.mixins.json import DataClassJSONMixin

from .entities import (
    MyDataClass,
    MyDatetimeNewType,
    MyEnum,
    MyFlag,
    MyFrozenDataClass,
    MyGenericDataClass,
    MyGenericList,
    MyIntEnum,
    MyIntFlag,
    MyNamedTuple,
    MyNativeStrEnum,
    MyStrEnum,
    MyUntypedNamedTuple,
    T,
    TAny,
    TDefaultInt,
    TInt,
    TIntStr,
)

NoneType = type(None)

TMyDataClass = typing.TypeVar("TMyDataClass", bound=MyDataClass)


def test_is_init_var():
    assert is_init_var(InitVar[int])
    assert not is_init_var(int)


def test_is_literal_unsupported_python(mocker):
    mocker.patch("mashumaro.core.meta.helpers.PY_39", False)
    mocker.patch("mashumaro.core.meta.helpers.PY_310_MIN", False)
    assert not is_literal(typing_extensions.Literal[1])


def test_no_code_builder(mocker):
    mocker.patch(
        "mashumaro.mixins.dict.DataClassDictMixin.__init_subclass__",
        lambda: ...,
    )

    @dataclass
    class DataClass(DataClassDictMixin):
        pass

    assert DataClass.__pre_deserialize__({}) is None
    assert DataClass.__post_deserialize__(DataClass()) is None
    assert DataClass().__pre_serialize__() is None
    assert DataClass().__post_serialize__({}) is None


def test_get_class_that_defines_method():
    class A:
        def foo(self): ...

        @classmethod
        def bar(cls): ...

        def foobar(self): ...

    class B(A):
        def foobar(self): ...

    assert get_class_that_defines_method("foo", B) == A
    assert get_class_that_defines_method("bar", B) == A
    assert get_class_that_defines_method("foobar", B) == B


def test_get_class_that_defines_field():
    @dataclass
    class A:
        x: int
        y: int
        z: int

    @dataclass
    class B(A):
        y: float
        z: int

    assert get_class_that_defines_field("x", B) == A
    assert get_class_that_defines_field("y", B) == B
    assert get_class_that_defines_field("z", B) == B


def test_get_unknown_declared_hook():
    builder = CodeBuilder(object)
    assert builder.get_declared_hook("unknown_name") is None


def test_is_dataclass_dict_mixin():
    assert is_dataclass_dict_mixin(DataClassDictMixin)
    assert not is_dataclass_dict_mixin(DataClassJSONMixin)


def test_is_dataclass_dict_mixin_subclass():
    assert is_dataclass_dict_mixin_subclass(DataClassDictMixin)
    assert is_dataclass_dict_mixin_subclass(DataClassJSONMixin)
    assert is_dataclass_dict_mixin_subclass(MyDataClass)


def test_is_type_var_any():
    assert is_type_var_any(T)
    assert is_type_var_any(TAny)
    assert not is_type_var_any(TInt)
    assert not is_type_var_any(TDefaultInt)
    assert not is_type_var_any(typing.Any)
    assert not is_type_var_any(TMyDataClass)


def test_type_name():
    assert type_name(TAny) == "typing.Any"
    assert type_name(TInt) == "int"
    assert type_name(TDefaultInt) == "int"
    assert type_name(TMyDataClass) == "tests.entities.MyDataClass"
    assert type_name(TIntStr) == "typing.Union[int, str]"
    assert type_name(typing.List[TInt]) == "typing.List[int]"
    assert type_name(typing.Tuple[int]) == "typing.Tuple[int]"
    assert type_name(typing.Tuple[int, ...]) == "typing.Tuple[int, ...]"
    assert type_name(typing.Tuple[()]) == "typing.Tuple[()]"
    assert type_name(typing.Set[int]) == "typing.Set[int]"
    assert type_name(typing.FrozenSet[int]) == "typing.FrozenSet[int]"
    assert type_name(typing.Deque[int]) == "typing.Deque[int]"
    assert type_name(typing.Dict[int, int]) == "typing.Dict[int, int]"
    assert type_name(typing.Mapping[int, int]) == "typing.Mapping[int, int]"
    assert (
        type_name(typing.MutableMapping[int, int])
        == "typing.MutableMapping[int, int]"
    )
    assert type_name(typing.Counter[int]) == "typing.Counter[int]"
    assert type_name(typing.ChainMap[int, int]) == "typing.ChainMap[int, int]"
    assert type_name(typing.Sequence[int]) == "typing.Sequence[int]"
    assert type_name(typing.Union[int, str]) == "typing.Union[int, str]"
    assert (
        type_name(typing.Union[int, typing.Any])
        == "typing.Union[int, typing.Any]"
    )
    assert (
        type_name(typing.OrderedDict[int, int])
        == "typing.OrderedDict[int, int]"
    )
    assert (
        type_name(typing.DefaultDict[int, int])
        == "typing.DefaultDict[int, int]"
    )
    assert type_name(typing.Optional[int]) == "typing.Optional[int]"
    assert type_name(typing.Union[None, int]) == "typing.Optional[int]"
    assert type_name(typing.Union[int, None]) == "typing.Optional[int]"
    assert type_name(None) == "None"
    assert type_name(NoneType) == "NoneType"
    assert type_name(NoneType, none_type_as_none=True) == "None"
    assert type_name(typing.List[NoneType]) == "typing.List[NoneType]"
    assert (
        type_name(typing.Union[int, str, None])
        == "typing.Union[int, str, None]"
    )
    assert type_name(typing.Optional[NoneType]) == "NoneType"

    assert (
        type_name(types.MappingProxyType[int, int]) == "mappingproxy[int, int]"
    )
    if PY_310_MIN:
        assert type_name(int | None) == "typing.Optional[int]"
        assert type_name(None | int) == "typing.Optional[int]"
        assert type_name(int | str) == "typing.Union[int, str]"
    if PY_310_MIN:
        assert (
            type_name(MyDatetimeNewType) == "tests.entities.MyDatetimeNewType"
        )
    else:
        assert type_name(MyDatetimeNewType) == type_name(datetime)
    assert (
        type_name(typing_extensions.Annotated[TMyDataClass, None])
        == "tests.entities.MyDataClass"
    )


def test_type_name_pep_585():
    assert type_name(list[str]) == "list[str]"
    assert type_name(collections.deque[str]) == "collections.deque[str]"
    assert type_name(tuple[str]) == "tuple[str]"
    assert type_name(tuple[str, ...]) == "tuple[str, ...]"
    assert type_name(tuple[()]) == "tuple[()]"
    assert type_name(set[str]) == "set[str]"
    assert type_name(frozenset[str]) == "frozenset[str]"
    assert type_name(collections.abc.Set[str]) == "collections.abc.Set[str]"
    assert (
        type_name(collections.abc.MutableSet[str])
        == "collections.abc.MutableSet[str]"
    )
    assert type_name(collections.Counter[str]) == "collections.Counter[str]"
    assert (
        type_name(collections.abc.Sequence[str])
        == "collections.abc.Sequence[str]"
    )
    assert (
        type_name(collections.abc.MutableSequence[str])
        == "collections.abc.MutableSequence[str]"
    )
    assert (
        type_name(collections.ChainMap[str, str])
        == "collections.ChainMap[str, str]"
    )
    assert type_name(dict[str, str]) == "dict[str, str]"
    assert (
        type_name(collections.abc.Mapping[str, str])
        == "collections.abc.Mapping[str, str]"
    )
    assert (
        type_name(collections.OrderedDict[str, str])
        == "collections.OrderedDict[str, str]"
    )
    assert (
        type_name(collections.defaultdict[str, str])
        == "collections.defaultdict[str, str]"
    )


def test_type_name_short():
    assert type_name(TAny, short=True) == "Any"
    assert type_name(TInt, short=True) == "int"
    assert type_name(TDefaultInt, short=True) == "int"
    assert type_name(TMyDataClass, short=True) == "MyDataClass"
    assert type_name(TIntStr, short=True) == "Union[int, str]"
    assert type_name(typing.List[TInt], short=True) == "List[int]"
    assert type_name(typing.Tuple[int], short=True) == "Tuple[int]"
    assert type_name(typing.Tuple[int, ...], short=True) == "Tuple[int, ...]"
    assert type_name(typing.Tuple[()], short=True) == "Tuple[()]"
    assert type_name(typing.Set[int], short=True) == "Set[int]"
    assert type_name(typing.FrozenSet[int], short=True) == "FrozenSet[int]"
    assert type_name(typing.Deque[int], short=True) == "Deque[int]"
    assert type_name(typing.Dict[int, int], short=True) == "Dict[int, int]"
    assert (
        type_name(typing.Mapping[int, int], short=True) == "Mapping[int, int]"
    )
    assert (
        type_name(typing.MutableMapping[int, int], short=True)
        == "MutableMapping[int, int]"
    )
    assert type_name(typing.Counter[int], short=True) == "Counter[int]"
    assert (
        type_name(typing.ChainMap[int, int], short=True)
        == "ChainMap[int, int]"
    )
    assert type_name(typing.Sequence[int], short=True) == "Sequence[int]"
    assert type_name(typing.Union[int, str], short=True) == "Union[int, str]"
    assert (
        type_name(typing.Union[int, typing.Any], short=True)
        == "Union[int, Any]"
    )
    assert (
        type_name(typing.OrderedDict[int, int], short=True)
        == "OrderedDict[int, int]"
    )
    assert (
        type_name(typing.DefaultDict[int, int], short=True)
        == "DefaultDict[int, int]"
    )
    assert type_name(typing.Optional[int], short=True) == "Optional[int]"
    assert type_name(typing.Union[None, int], short=True) == "Optional[int]"
    assert type_name(typing.Union[int, None], short=True) == "Optional[int]"
    assert type_name(None, short=True) == "None"
    assert type_name(NoneType, short=True) == "NoneType"
    assert type_name(NoneType, short=True, none_type_as_none=True) == "None"
    assert type_name(typing.List[NoneType], short=True) == "List[NoneType]"
    assert (
        type_name(typing.Union[int, str, None], short=True)
        == "Union[int, str, None]"
    )
    assert type_name(typing.Optional[NoneType], short=True) == "NoneType"

    assert (
        type_name(types.MappingProxyType[int, int], short=True)
        == "mappingproxy[int, int]"
    )
    if PY_310_MIN:
        assert type_name(int | None, short=True) == "Optional[int]"
        assert type_name(None | int, short=True) == "Optional[int]"
        assert type_name(int | str, short=True) == "Union[int, str]"
    if PY_310_MIN:
        assert type_name(MyDatetimeNewType, short=True) == "MyDatetimeNewType"
    else:
        assert type_name(MyDatetimeNewType, short=True) == type_name(
            datetime, short=True
        )
    assert (
        type_name(typing_extensions.Annotated[TMyDataClass, None], short=True)
        == "MyDataClass"
    )


def test_type_name_pep_585_short():
    assert type_name(list[str], short=True) == "list[str]"
    assert type_name(collections.deque[str], short=True) == "deque[str]"
    assert type_name(tuple[str], short=True) == "tuple[str]"
    assert type_name(tuple[str, ...], short=True) == "tuple[str, ...]"
    assert type_name(tuple[()], short=True) == "tuple[()]"
    assert type_name(set[str], short=True) == "set[str]"
    assert type_name(frozenset[str], short=True) == "frozenset[str]"
    assert type_name(collections.abc.Set[str], short=True) == "Set[str]"
    assert (
        type_name(collections.abc.MutableSet[str], short=True)
        == "MutableSet[str]"
    )
    assert type_name(collections.Counter[str], short=True) == "Counter[str]"
    assert (
        type_name(collections.abc.Sequence[str], short=True) == "Sequence[str]"
    )
    assert (
        type_name(collections.abc.MutableSequence[str], short=True)
        == "MutableSequence[str]"
    )
    assert (
        type_name(collections.ChainMap[str, str], short=True)
        == "ChainMap[str, str]"
    )
    assert type_name(dict[str, str], short=True) == "dict[str, str]"
    assert (
        type_name(collections.abc.Mapping[str, str], short=True)
        == "Mapping[str, str]"
    )
    assert (
        type_name(collections.OrderedDict[str, str], short=True)
        == "OrderedDict[str, str]"
    )
    assert (
        type_name(collections.defaultdict[str, str], short=True)
        == "defaultdict[str, str]"
    )


def test_get_type_origin():
    assert get_type_origin(typing.List[int]) == list
    assert get_type_origin(typing.List) == list
    assert get_type_origin(MyGenericDataClass[int]) == MyGenericDataClass
    assert get_type_origin(MyGenericDataClass) == MyGenericDataClass
    assert (
        get_type_origin(typing_extensions.Annotated[datetime, None])
        == datetime
    )
    assert (
        get_type_origin(typing_extensions.Required[int])
        == typing_extensions.Required
    )


def test_resolve_type_params():
    @dataclass
    class A(typing.Generic[T]):
        x: T

    @dataclass
    class B(A[int]):
        pass

    resolved = resolve_type_params(B)
    assert resolved[A] == {T: int}
    assert resolved[B] == {}


def test_get_generic_name():
    assert get_generic_name(typing.List[int]) == "typing.List"
    assert get_generic_name(typing.List[int], short=True) == "List"
    assert (
        get_generic_name(MyGenericDataClass[int])
        == "tests.entities.MyGenericDataClass"
    )
    assert (
        get_generic_name(MyGenericDataClass[int], short=True)
        == "MyGenericDataClass"
    )


def test_get_generic_collection_based_class_name():
    assert get_generic_name(MyGenericList, short=True) == "MyGenericList"
    assert get_generic_name(MyGenericList) == "tests.entities.MyGenericList"
    assert get_generic_name(MyGenericList[int], short=True) == "MyGenericList"
    assert (
        get_generic_name(MyGenericList[int]) == "tests.entities.MyGenericList"
    )


def test_is_dialect_subclass():
    class MyDialect(Dialect):
        pass

    assert is_dialect_subclass(Dialect)
    assert is_dialect_subclass(MyDialect)
    assert not is_dialect_subclass(123)


def test_is_union():
    t = typing.Optional[str]
    assert is_union(t)
    assert get_args(t) == (str, NoneType)
    t = typing.Union[str, None]
    assert is_union(t)
    assert get_args(t) == (str, NoneType)
    t = typing.Union[None, str]
    assert is_union(t)
    assert get_args(t) == (NoneType, str)


@pytest.mark.skipif(not PY_310_MIN, reason="requires python 3.10+")
def test_is_union_pep_604():
    t = str | None
    assert is_union(t)
    assert get_args(t) == (str, NoneType)
    t = None | str
    assert is_union(t)
    assert get_args(t) == (NoneType, str)


def test_is_optional():
    t = typing.Optional[str]
    assert is_optional(t)
    assert get_args(t) == (str, NoneType)
    t = typing.Union[str, None]
    assert is_optional(t)
    assert get_args(t) == (str, NoneType)
    t = typing.Union[None, str]
    assert is_optional(t)
    assert get_args(t) == (NoneType, str)


@pytest.mark.skipif(not PY_310_MIN, reason="requires python 3.10+")
def test_is_optional_pep_604():
    t = str | None
    assert is_optional(t)
    assert get_args(t) == (str, NoneType)
    t = None | str
    assert is_optional(t)
    assert get_args(t) == (NoneType, str)


def test_not_non_type_arg():
    assert not_none_type_arg((str, int)) == str
    assert not_none_type_arg((NoneType, int)) == int
    assert not_none_type_arg((str, NoneType)) == str
    assert not_none_type_arg((T, int), {T: NoneType}) == int
    assert not_none_type_arg((NoneType,)) is None


def test_is_named_tuple():
    assert is_named_tuple(MyNamedTuple)
    assert is_named_tuple(MyUntypedNamedTuple)
    assert not is_named_tuple(object())


def test_is_new_type():
    assert is_new_type(typing.NewType("MyNewType", int))
    assert not is_new_type(int)


def test_is_annotated():
    assert is_annotated(typing_extensions.Annotated[datetime, None])
    assert not is_annotated(datetime)


def test_is_literal():
    assert is_literal(typing_extensions.Literal[1, 2, 3])
    assert not is_literal(typing_extensions.Literal)
    assert not is_literal([1, 2, 3])


def test_get_literal_values():
    assert get_literal_values(typing_extensions.Literal[1, 2, 3]) == (1, 2, 3)
    assert get_literal_values(
        typing_extensions.Literal[
            1, typing_extensions.Literal[typing_extensions.Literal[2], 3]
        ]
    ) == (1, 2, 3)


def test_type_name_literal():
    assert type_name(
        getattr(typing, "Literal")[
            1,
            "a",
            b"\x00",
            True,
            False,
            None,
            MyEnum.a,
            MyStrEnum.a,
            MyNativeStrEnum.a,
            MyIntEnum.a,
            MyFlag.a,
            MyIntFlag.a,
            typing_extensions.Literal[2, 3],
            typing_extensions.Literal[typing_extensions.Literal["b", "c"]],
        ]
    ) == (
        f"typing.Literal[1, 'a', b'\\x00', True, False, None, "
        "tests.entities.MyEnum.a, tests.entities.MyStrEnum.a, "
        "tests.entities.MyNativeStrEnum.a, tests.entities.MyIntEnum.a, "
        "tests.entities.MyFlag.a, tests.entities.MyIntFlag.a, 2, 3, 'b', 'c']"
    )


def test_code_builder_get_pack_method_name():
    builder = CodeBuilder(object)
    type_args = (int,)
    type_args_hash = hash_type_args((int,))

    method_name = builder.get_pack_method_name()
    assert method_name == "__mashumaro_to_dict__"
    assert method_name.public == "to_dict"

    method_name = builder.get_pack_method_name(type_args=type_args)
    assert method_name == f"__mashumaro_to_dict_{type_args_hash}__"
    assert method_name.public == f"to_dict_{type_args_hash}"

    method_name = builder.get_pack_method_name(
        type_args=type_args, format_name="yaml"
    )
    assert method_name == f"__mashumaro_to_dict_yaml_{type_args_hash}__"
    assert method_name.public == f"to_dict_yaml_{type_args_hash}"

    method_name = builder.get_pack_method_name(
        type_args=type_args, format_name="yaml"
    )
    assert method_name == f"__mashumaro_to_dict_yaml_{type_args_hash}__"
    assert method_name.public == f"to_dict_yaml_{type_args_hash}"

    method_name = builder.get_pack_method_name(format_name="yaml")
    assert method_name == "__mashumaro_to_dict_yaml__"
    assert method_name.public == "to_dict_yaml"

    method_name = builder.get_pack_method_name(
        format_name="yaml", encoder=object
    )
    assert method_name == "__mashumaro_to_yaml__"
    assert method_name.public == "to_yaml"

    method_name = builder.get_pack_method_name(
        type_args=type_args, encoder=object
    )
    assert method_name == f"__mashumaro_to_dict_{type_args_hash}__"
    assert method_name.public == f"to_dict_{type_args_hash}"

    method_name = builder.get_pack_method_name(encoder=object)
    assert method_name == "__mashumaro_to_dict__"
    assert method_name.public == "to_dict"

    method_name = builder.get_pack_method_name(
        type_args=type_args, format_name="yaml", encoder=object
    )
    assert method_name == "__mashumaro_to_yaml__"
    assert method_name.public == "to_yaml"


def test_code_builder_get_unpack_method_name():
    builder = CodeBuilder(object)
    type_args = (int,)
    type_args_hash = hash_type_args((int,))

    method_name = builder.get_unpack_method_name()
    assert method_name == "__mashumaro_from_dict__"
    assert method_name.public == "from_dict"

    method_name = builder.get_unpack_method_name(type_args=type_args)
    assert method_name == f"__mashumaro_from_dict_{type_args_hash}__"
    assert method_name.public == f"from_dict_{type_args_hash}"

    method_name = builder.get_unpack_method_name(
        type_args=type_args, format_name="yaml"
    )
    assert method_name == f"__mashumaro_from_dict_yaml_{type_args_hash}__"
    assert method_name.public == f"from_dict_yaml_{type_args_hash}"

    method_name = builder.get_unpack_method_name(format_name="yaml")
    assert method_name == "__mashumaro_from_dict_yaml__"
    assert method_name.public == "from_dict_yaml"

    method_name = builder.get_unpack_method_name(
        format_name="yaml", decoder=object
    )
    assert method_name == "__mashumaro_from_yaml__"
    assert method_name.public == "from_yaml"

    method_name = builder.get_unpack_method_name(
        type_args=type_args, decoder=object
    )
    assert method_name == f"__mashumaro_from_dict_{type_args_hash}__"
    assert method_name.public == f"from_dict_{type_args_hash}"

    method_name = builder.get_unpack_method_name(decoder=object)
    assert method_name == "__mashumaro_from_dict__"
    assert method_name.public == "from_dict"

    method_name = builder.get_unpack_method_name(
        type_args=type_args, format_name="yaml", decoder=object
    )
    assert method_name == "__mashumaro_from_yaml__"
    assert method_name.public == "from_yaml"


def test_is_self():
    assert is_self(typing_extensions.Self)
    assert not is_self(object)


def test_ensure_generic_collection_not_generic():
    assert not ensure_generic_collection(
        ValueSpec(int, "", CodeBuilder(None), FieldContext("", {}))
    )


def test_ensure_mapping_key_type_hashable():
    spec = ValueSpec(typing.Dict, "", CodeBuilder(None), FieldContext("", {}))

    with pytest.raises(UnserializableField):
        ensure_mapping_key_type_hashable(spec, (dict, str))
    with pytest.raises(UnserializableField):
        ensure_mapping_key_type_hashable(spec, (MyDataClass, str))
    assert ensure_mapping_key_type_hashable(spec, (MyFrozenDataClass, str))


def test_get_function_arg_annotation():
    def foo(x: int, y: Dialect) -> None: ...

    assert get_function_arg_annotation(foo, "x") == int
    assert get_function_arg_annotation(foo, "y") == Dialect
    assert get_function_arg_annotation(foo, arg_name="x") == int
    assert get_function_arg_annotation(foo, arg_name="y") == Dialect
    assert get_function_arg_annotation(foo, arg_pos=0) == int
    assert get_function_arg_annotation(foo, arg_pos=1) == Dialect
    with pytest.raises(ValueError):
        get_function_arg_annotation(foo)


def test_get_function_return_annotation():
    def foo(x: int, y: Dialect) -> Dialect: ...

    assert get_function_return_annotation(foo) == Dialect


def test_collect_type_params():
    T = typing.TypeVar("T")
    S = typing.TypeVar("S")

    class MyGeneric(typing.Generic[T, S], typing.Mapping[T, S]):
        pass

    assert collect_type_params(MyGeneric[T, T]) == [T]


def test_is_generic_like_with_class_getitem():
    class MyClass:
        def __class_getitem__(cls, item):
            return cls

    assert is_generic(MyClass)
    assert is_generic(MyClass[int])


def test_get_type_annotations():
    assert get_type_annotations(int) == []
    assert get_type_annotations(typing_extensions.Annotated[int, 42]) == (42,)


def test_substitute_type_params():
    assert substitute_type_params(int, {}) == int
    assert substitute_type_params(T, {T: int}) == int
    assert (
        substitute_type_params(typing.Dict[T, TAny], {T: str})
        == typing.Dict[str, TAny]
    )
    assert (
        substitute_type_params(typing_extensions.Annotated[T, 42], {T: int})
        == typing_extensions.Annotated[int, 42]
    )
    assert (
        substitute_type_params(
            typing.Dict[str, typing_extensions.Annotated[int, {}]], {T: int}
        )
        == typing.Dict[str, typing_extensions.Annotated[int, {}]]
    )


def test_is_hashable():
    assert is_hashable(42) is True
    assert is_hashable({}) is False
    assert is_hashable(typing_extensions.Annotated[int, 42]) is True
    assert is_hashable(typing_extensions.Annotated[int, {}]) is False


def test_is_hashable_type():
    assert is_hashable_type(dict) is False
    assert is_hashable_type(int) is True
    assert is_hashable_type(MyFrozenDataClass) is True
    assert is_hashable_type(MyDataClass) is False


def test_type_var_has_default():
    T_WithoutDefault = typing_extensions.TypeVar("T_WithoutDefault")
    T_WithDefault = typing_extensions.TypeVar("T_WithDefault", default=int)
    T_WithDefaultNone = typing_extensions.TypeVar(
        "T_WithDefaultNone", default=None
    )

    assert not type_var_has_default(T_WithoutDefault)
    assert type_var_has_default(T_WithDefault)
    assert type_var_has_default(T_WithDefaultNone)


def test_get_type_var_default():
    T_WithDefault = typing_extensions.TypeVar("T_WithDefault", default=int)
    T_WithDefaultNone = typing_extensions.TypeVar(
        "T_WithDefaultNone", default=None
    )

    assert get_type_var_default(T_WithDefault) is int
    assert get_type_var_default(T_WithDefaultNone) is None



from dataclasses import dataclass
from typing import List

import yaml

from mashumaro.mixins.yaml import DataClassYAMLMixin


def test_to_yaml():
    @dataclass
    class DataClass(DataClassYAMLMixin):
        x: List[int]

    dumped = yaml.dump({"x": [1, 2, 3]})
    assert DataClass([1, 2, 3]).to_yaml() == dumped


def test_from_yaml():
    @dataclass
    class DataClass(DataClassYAMLMixin):
        x: List[int]

    dumped = yaml.dump({"x": [1, 2, 3]})
    assert DataClass.from_yaml(dumped) == DataClass([1, 2, 3])



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional

import msgpack
import orjson
import pytest

from mashumaro import DataClassDictMixin
from mashumaro.config import ADD_DIALECT_SUPPORT, BaseConfig
from mashumaro.dialect import Dialect
from mashumaro.exceptions import UnresolvedTypeReferenceError
from mashumaro.mixins.msgpack import DataClassMessagePackMixin
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.types import SerializationStrategy

from .conftest import add_unpack_method


@dataclass
class A(DataClassDictMixin):
    x: B


@dataclass
class B(DataClassDictMixin):
    x: int


@dataclass
class Base(DataClassDictMixin):
    pass


@dataclass
class A1(Base):
    a: B1


@dataclass
class A2(Base):
    a: B2


@dataclass
class A3(Base):
    a: B1
    x: int

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


@dataclass
class B1(Base):
    b: int


@dataclass
class AMessagePack(DataClassMessagePackMixin):
    x: BMessagePack


@dataclass
class BMessagePack(DataClassMessagePackMixin):
    x: int


@dataclass
class BaseMessagePack(DataClassMessagePackMixin):
    pass


@dataclass
class A1MessagePack(BaseMessagePack):
    a: B1MessagePack


@dataclass
class A2MessagePack(BaseMessagePack):
    a: B2


@dataclass
class A3MessagePack(BaseMessagePack):
    a: B1MessagePack
    x: int

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


@dataclass
class B1MessagePack(BaseMessagePack):
    b: int


@dataclass
class A3ORJSON(DataClassORJSONMixin):
    a: B1ORJSON
    x: int

    class Config(BaseConfig):
        code_generation_options = [ADD_DIALECT_SUPPORT]


@dataclass
class B1ORJSON(DataClassORJSONMixin):
    b: int


def test_postponed_annotation_evaluation():
    obj = A(x=B(x=1))
    assert obj.to_dict() == {"x": {"x": 1}}
    assert A.from_dict({"x": {"x": 1}}) == obj


def test_unresolved_type_with_allowed_postponed_annotation_evaluation():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: X

    with pytest.raises(UnresolvedTypeReferenceError):
        DataClass.from_dict({})

    with pytest.raises(UnresolvedTypeReferenceError):
        DataClass(x=1).to_dict()


def test_unresolved_type_with_disallowed_postponed_annotation_evaluation():
    with pytest.raises(UnresolvedTypeReferenceError):

        @dataclass
        class DataClass(DataClassDictMixin):
            x: X

            class Config(BaseConfig):
                allow_postponed_evaluation = False

    with add_unpack_method:
        with pytest.raises(UnresolvedTypeReferenceError):

            @dataclass
            class DataClass(DataClassDictMixin):
                x: X

                class Config(BaseConfig):
                    allow_postponed_evaluation = False


def test_postponed_annotation_evaluation_with_parent():
    obj = A1(B1(1))
    assert A1.from_dict({"a": {"b": 1}}) == obj
    assert obj.to_dict() == {"a": {"b": 1}}


def test_postponed_annotation_evaluation_with_parent_and_no_reference():
    with pytest.raises(UnresolvedTypeReferenceError):
        A2.from_dict({"a": {"b": 1}})
    with pytest.raises(UnresolvedTypeReferenceError):
        A2(None).to_dict()


def test_postponed_annotation_evaluation_with_parent_and_dialect():
    class MyDialect(Dialect):
        serialization_strategy = {
            int: {
                "serialize": lambda i: str(int(i * 1000)),
                "deserialize": lambda s: int(int(s) / 1000),
            }
        }

    obj = A3(B1(1), 2)
    assert A3.from_dict({"a": {"b": 1}, "x": 2}) == obj
    assert A3.from_dict({"a": {"b": 1}, "x": "2000"}, dialect=MyDialect) == obj
    assert obj.to_dict() == {"a": {"b": 1}, "x": 2}
    assert obj.to_dict(dialect=MyDialect) == {"a": {"b": 1}, "x": "2000"}


def test_postponed_annotation_evaluation_msgpack():
    obj = AMessagePack(x=BMessagePack(x=1))
    assert obj.to_dict() == {"x": {"x": 1}}
    assert AMessagePack.from_dict({"x": {"x": 1}}) == obj
    dump = msgpack.dumps({"x": {"x": 1}})
    assert obj.to_msgpack() == dump
    assert AMessagePack.from_msgpack(dump) == obj


def test_unresolved_type_with_allowed_postponed_annotation_evaluation_msgpack():
    @dataclass
    class DataClass(DataClassMessagePackMixin):
        x: X

    with pytest.raises(UnresolvedTypeReferenceError):
        DataClass.from_msgpack(b"")

    with pytest.raises(UnresolvedTypeReferenceError):
        DataClass(x=1).to_msgpack()


def test_postponed_annotation_evaluation_with_parent_msgpack():
    obj = A1MessagePack(B1MessagePack(1))
    dump = msgpack.dumps({"a": {"b": 1}})
    assert A1MessagePack.from_msgpack(dump) == obj
    assert obj.to_msgpack() == dump


def test_postponed_annotation_evaluation_with_parent_and_no_reference_msgpack():
    with pytest.raises(UnresolvedTypeReferenceError):
        A2MessagePack.from_msgpack(b"")
    with pytest.raises(UnresolvedTypeReferenceError):
        A2MessagePack(None).to_msgpack()


def test_postponed_annotation_evaluation_with_parent_and_dialect_msgpack():
    class MyDialect(Dialect):
        serialization_strategy = {
            int: {
                "serialize": lambda i: str(int(i * 1000)),
                "deserialize": lambda s: int(int(s) / 1000),
            }
        }

    obj = A3MessagePack(B1MessagePack(1), 2)
    dump = msgpack.dumps({"a": {"b": 1}, "x": 2})
    dump_dialect = msgpack.dumps({"a": {"b": 1}, "x": "2000"})
    assert A3MessagePack.from_msgpack(dump) == obj
    assert A3MessagePack.from_msgpack(dump_dialect, dialect=MyDialect) == obj
    assert obj.to_msgpack() == dump
    assert obj.to_msgpack(dialect=MyDialect) == dump_dialect


def test_postponed_msgpack_with_custom_encoder_and_decoder():
    def decoder(data) -> Dict[str, bytes]:
        def modify(d):
            result = {}
            for k, v in d.items():
                if isinstance(v, dict):
                    result[k] = modify(v)
                else:
                    result[k] = v // 1000
            return result

        return modify(msgpack.loads(data))

    def encoder(data: Dict[str, bytes]) -> bytes:
        def modify(d):
            result = {}
            for k, v in d.items():
                if isinstance(v, dict):
                    result[k] = modify(v)
                else:
                    result[k] = v * 1000
            return result

        return msgpack.dumps(modify(data))

    instance = A3MessagePack(B1MessagePack(123), 456)
    dumped = msgpack.packb({"a": {"b": 123000}, "x": 456000})
    assert instance.to_msgpack(encoder=encoder) == dumped
    assert A3MessagePack.from_msgpack(dumped, decoder=decoder) == instance


def test_postponed_orjson_with_custom_encoder_and_decoder():
    def decoder(data) -> Dict[str, bytes]:
        def modify(d):
            result = {}
            for k, v in d.items():
                if isinstance(v, dict):
                    result[k] = modify(v)
                else:
                    result[k] = v // 1000
            return result

        return modify(orjson.loads(data))

    def encoder(data: Dict[str, bytes], **_) -> bytes:
        def modify(d):
            result = {}
            for k, v in d.items():
                if isinstance(v, dict):
                    result[k] = modify(v)
                else:
                    result[k] = v * 1000
            return result

        return orjson.dumps(modify(data))

    instance = A3ORJSON(B1ORJSON(123), 456)
    dumped = orjson.dumps({"a": {"b": 123000}, "x": 456000})
    assert instance.to_jsonb(encoder=encoder) == dumped
    assert A3ORJSON.from_json(dumped, decoder=decoder) == instance


def test_postponed_serialization_strategy() -> None:
    class Strategy(SerializationStrategy, use_annotations=True):
        def serialize(self, value) -> dict[str, Any]:
            return {"a": value}

        def deserialize(self, value: dict[str, Any]):
            return value.get("a")

    @dataclass
    class MyDataClass(DataClassDictMixin):
        x: Optional[int]

        class Config(BaseConfig):
            serialization_strategy = {int: Strategy()}

    obj = MyDataClass(x=2)
    assert obj.to_dict() == {"x": {"a": 2}}
    assert MyDataClass.from_dict({"x": {"a": 2}}) == obj



from dataclasses import dataclass
from datetime import date, datetime
from typing import Generic, List, Optional, TypeVar, Union

import pytest
from typing_extensions import Literal

from mashumaro.codecs import BasicDecoder, BasicEncoder
from mashumaro.codecs.basic import decode, encode
from mashumaro.dialect import Dialect
from tests.entities import (
    DataClassWithoutMixin,
    GenericNamedTuple,
    GenericTypedDict,
    MyGenericDataClass,
)

T = TypeVar("T")


@dataclass
class Foo:
    foo: str


@dataclass
class Bar:
    bar: str


class MyDialect(Dialect):
    serialization_strategy = {
        date: {
            "serialize": date.toordinal,
            "deserialize": date.fromordinal,
        },
    }


@dataclass
class GenericDataClass(Generic[T]):
    x: T
    y: List[T]


@pytest.mark.parametrize(
    ("shape_type", "encoded", "decoded"),
    [
        [
            List[date],
            ["2023-09-22", "2023-09-23"],
            [
                date(2023, 9, 22),
                date(2023, 9, 23),
            ],
        ],
        [DataClassWithoutMixin, {"i": 42}, DataClassWithoutMixin(42)],
        [
            List[DataClassWithoutMixin],
            [{"i": 42}],
            [DataClassWithoutMixin(42)],
        ],
        [
            GenericDataClass,
            {"x": "2023-09-23", "y": ["2023-09-23"]},
            GenericDataClass("2023-09-23", ["2023-09-23"]),
        ],
        [
            GenericDataClass[date],
            {"x": "2023-09-23", "y": ["2023-09-23"]},
            GenericDataClass(date(2023, 9, 23), [date(2023, 9, 23)]),
        ],
        [
            List[Union[int, date]],
            ["42", "2023-09-23"],
            [42, date(2023, 9, 23)],
        ],
        [Optional[date], "2023-09-23", date(2023, 9, 23)],
        [Optional[date], None, None],
    ],
)
def test_decode(shape_type, encoded, decoded):
    assert decode(encoded, shape_type) == decoded


@pytest.mark.parametrize(
    ("shape_type", "encoded", "decoded"),
    [
        [
            List[date],
            ["2023-09-22", "2023-09-23"],
            [
                date(2023, 9, 22),
                date(2023, 9, 23),
            ],
        ],
        [DataClassWithoutMixin, {"i": 42}, DataClassWithoutMixin(42)],
        [
            List[DataClassWithoutMixin],
            [{"i": 42}],
            [DataClassWithoutMixin(42)],
        ],
        [
            GenericDataClass,
            {"x": date(2023, 9, 23), "y": [date(2023, 9, 23)]},
            GenericDataClass(date(2023, 9, 23), [date(2023, 9, 23)]),
        ],
        [
            GenericDataClass[date],
            {"x": "2023-09-23", "y": ["2023-09-23"]},
            GenericDataClass(date(2023, 9, 23), [date(2023, 9, 23)]),
        ],
        [List[Union[int, date]], [42, "2023-09-23"], [42, date(2023, 9, 23)]],
        [Optional[date], "2023-09-23", date(2023, 9, 23)],
        [Optional[date], None, None],
    ],
)
def test_encode(shape_type, encoded, decoded):
    assert encode(decoded, shape_type) == encoded


def test_decoder_with_default_dialect():
    decoder = BasicDecoder(List[date], default_dialect=MyDialect)
    assert decoder.decode([738785, 738786]) == [
        date(2023, 9, 22),
        date(2023, 9, 23),
    ]


def test_encoder_with_default_dialect():
    encoder = BasicEncoder(List[date], default_dialect=MyDialect)
    assert encoder.encode([date(2023, 9, 22), date(2023, 9, 23)]) == [
        738785,
        738786,
    ]


def test_pre_decoder_func():
    decoder = BasicDecoder(List[date], pre_decoder_func=lambda v: v.split(","))
    assert decoder.decode("2023-09-22,2023-09-23") == [
        date(2023, 9, 22),
        date(2023, 9, 23),
    ]


def test_post_encoder_func():
    encoder = BasicEncoder(List[date], post_encoder_func=lambda v: ",".join(v))
    assert (
        encoder.encode(
            [
                date(2023, 9, 22),
                date(2023, 9, 23),
            ]
        )
        == "2023-09-22,2023-09-23"
    )


@pytest.mark.parametrize(
    ("shape_type", "invalid_value"),
    [[Union[date, datetime], "foo"], [Literal["foo"], "bar"]],
)
def test_value_error_on_decode(shape_type, invalid_value):
    decoder = BasicDecoder(shape_type)
    with pytest.raises(ValueError) as e:
        decoder.decode(invalid_value)
    assert type(e.value) is ValueError


@pytest.mark.parametrize(
    ("shape_type", "invalid_value"),
    [[Union[date, datetime], "foo"], [Literal["foo"], "bar"]],
)
def test_value_error_on_encode(shape_type, invalid_value):
    encoder = BasicEncoder(shape_type)
    with pytest.raises(ValueError) as e:
        encoder.encode(invalid_value)
    assert type(e.value) is ValueError


def test_with_fields_with_generated_methods():
    @dataclass
    class MyClass:
        td1: GenericTypedDict[str]
        td2: GenericTypedDict[date]
        nt1: GenericNamedTuple[str]
        nt2: GenericNamedTuple[date]
        u1: List[Union[int, str]]
        u2: List[Union[int, date]]
        l1: Literal["l1"]
        l2: Literal["l2"]

    decoder = BasicDecoder(MyClass)
    encoder = BasicEncoder(MyClass)
    data = {
        "td1": {"x": "2023-11-15", "y": 1},
        "td2": {"x": "2023-11-15", "y": 2},
        "nt1": ["2023-11-15", 3],
        "nt2": ["2023-11-15", 4],
        "u1": [5, "2023-11-15"],
        "u2": [6, "2023-11-15"],
        "l1": "l1",
        "l2": "l2",
    }
    obj = MyClass(
        td1={"x": "2023-11-15", "y": 1},
        td2={"x": date(2023, 11, 15), "y": 2},
        nt1=GenericNamedTuple("2023-11-15", 3),
        nt2=GenericNamedTuple(date(2023, 11, 15), 4),
        u1=[5, "2023-11-15"],
        u2=[6, date(2023, 11, 15)],
        l1="l1",
        l2="l2",
    )
    assert decoder.decode(data) == obj
    assert encoder.encode(obj) == data


def test_with_two_dataclass_fields():
    @dataclass
    class MyClass:
        x1: Foo
        x2: Bar

    decoder = BasicDecoder(MyClass)
    encoder = BasicEncoder(MyClass)
    data = {"x1": {"foo": "foo"}, "x2": {"bar": "bar"}}
    obj = MyClass(x1=Foo("foo"), x2=Bar("bar"))
    assert decoder.decode(data) == obj
    assert encoder.encode(obj) == data


def test_with_two_generic_dataclass_fields():
    @dataclass
    class MyClass:
        x1: MyGenericDataClass[str]
        x2: MyGenericDataClass[date]

    decoder = BasicDecoder(MyClass)
    encoder = BasicEncoder(MyClass)
    data = {"x1": {"x": "2023-11-15"}, "x2": {"x": "2023-11-15"}}
    obj = MyClass(
        x1=MyGenericDataClass("2023-11-15"),
        x2=MyGenericDataClass(date(2023, 11, 15)),
    )
    assert decoder.decode(data) == obj
    assert encoder.encode(obj) == data



from dataclasses import dataclass
from datetime import timedelta, timezone

import pytest

from mashumaro import DataClassDictMixin

time_zones = [
    (timezone(timedelta(hours=-12)), "UTC-12:00"),  # Baker Island
    (timezone(timedelta(hours=-11)), "UTC-11:00"),  # American Samoa
    (timezone(timedelta(hours=-10)), "UTC-10:00"),  # Hawaii
    (
        timezone(timedelta(hours=-9, minutes=-30)),
        "UTC-09:30",
    ),  # Marquesas Isl.
    (timezone(timedelta(hours=-9)), "UTC-09:00"),  # Gambier Isl.
    (timezone(timedelta(hours=-8)), "UTC-08:00"),  # Pitcairn Isl.
    (timezone(timedelta(hours=-7)), "UTC-07:00"),  # Sonora
    (timezone(timedelta(hours=-6)), "UTC-06:00"),  # Costa Rica
    (timezone(timedelta(hours=-5)), "UTC-05:00"),  # Colombia
    (timezone(timedelta(hours=-4)), "UTC-04:00"),  # Bolivia
    (timezone(timedelta(hours=-3, minutes=-30)), "UTC-03:30"),  # Canada
    (timezone(timedelta(hours=-3)), "UTC-03:00"),  # Argentina
    (timezone(timedelta(hours=-2)), "UTC-02:00"),  # South Georgia
    (timezone(timedelta(hours=-1)), "UTC-01:00"),  # Cape Verde
    (timezone(timedelta(hours=0)), "UTC"),  # Burkina Faso
    (timezone(timedelta(hours=1)), "UTC+01:00"),  # Algeria
    (timezone(timedelta(hours=2)), "UTC+02:00"),  # Botswana
    (timezone(timedelta(hours=3)), "UTC+03:00"),  # Moscow
    (timezone(timedelta(hours=3, minutes=30)), "UTC+03:30"),  # Iran
    (timezone(timedelta(hours=4)), "UTC+04:00"),  # Armenia
    (timezone(timedelta(hours=4, minutes=30)), "UTC+04:30"),  # Afghanistan
    (timezone(timedelta(hours=5)), "UTC+05:00"),  # Maldives
    (timezone(timedelta(hours=5, minutes=30)), "UTC+05:30"),  # India
    (timezone(timedelta(hours=5, minutes=45)), "UTC+05:45"),  # Nepal
    (timezone(timedelta(hours=6)), "UTC+06:00"),  # Bangladesh
    (timezone(timedelta(hours=6, minutes=30)), "UTC+06:30"),  # Myanmar
    (timezone(timedelta(hours=7)), "UTC+07:00"),  # Cambodia
    (timezone(timedelta(hours=8)), "UTC+08:00"),  # Hong Kong
    (timezone(timedelta(hours=8, minutes=45)), "UTC+08:45"),  # Eucla
    (timezone(timedelta(hours=9)), "UTC+09:00"),  # Japan
    (timezone(timedelta(hours=9, minutes=30)), "UTC+09:30"),  # N. Australia
    (timezone(timedelta(hours=10)), "UTC+10:00"),  # Queensland
    (timezone(timedelta(hours=10, minutes=30)), "UTC+10:30"),  # Australia
    (timezone(timedelta(hours=11)), "UTC+11:00"),  # Vanuatu
    (timezone(timedelta(hours=12)), "UTC+12:00"),  # Nauru
    (timezone(timedelta(hours=12, minutes=45)), "UTC+12:45"),  # New Zealand
    (timezone(timedelta(hours=13)), "UTC+13:00"),  # Tonga
    (timezone(timedelta(hours=14)), "UTC+14:00"),  # Kiribati
]


@pytest.mark.parametrize(["tz", "tz_string"], time_zones)
def test_timezones(tz, tz_string):
    @dataclass
    class DataClass(DataClassDictMixin):
        x: timezone

    assert DataClass(tz).to_dict() == {"x": tz_string}
    assert DataClass.from_dict({"x": tz_string}) == DataClass(tz)


def test_invalid_timezone():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: timezone

    with pytest.raises(ValueError):
        DataClass.from_dict({"x": "UTC+03:00:01"})



from dataclasses import dataclass
from datetime import date

from mashumaro import DataClassDictMixin
from mashumaro.codecs import BasicDecoder, BasicEncoder


def test_type_alias_type_with_dataclass_dict_mixin():
    type MyDate = date

    @dataclass
    class MyClass(DataClassDictMixin):
        x: MyDate

    obj = MyClass(date(2024, 4, 15))
    assert MyClass.from_dict({"x": "2024-04-15"}) == obj
    assert obj.to_dict() == {"x": "2024-04-15"}


def test_type_alias_type_with_codecs():
    type MyDate = date
    decoder = BasicDecoder(MyDate)
    encoder = BasicEncoder(MyDate)

    obj = date(2024, 4, 15)
    assert decoder.decode("2024-04-15") == obj
    assert encoder.encode(obj) == "2024-04-15"



from dataclasses import dataclass
from datetime import date
from functools import partial
from typing import Generic, Tuple, TypeVar

try:
    from typing import TypeVarTuple
except ImportError:
    from typing_extensions import TypeVarTuple
try:
    from typing import Unpack
except ImportError:
    from typing_extensions import Unpack

import pytest

from mashumaro import DataClassDictMixin
from mashumaro.core.const import PY_311_MIN

# noinspection PyProtectedMember
from mashumaro.core.meta.helpers import (
    _check_generic,
    _flatten_type_args,
    resolve_type_params,
    type_name,
)
from mashumaro.core.meta.types.common import FieldContext, ValueSpec

# noinspection PyProtectedMember
from mashumaro.core.meta.types.pack import pack_tuple

# noinspection PyProtectedMember
from mashumaro.core.meta.types.unpack import unpack_tuple
from mashumaro.exceptions import MissingField

K = TypeVar("K")
V = TypeVar("V")
Ts = TypeVarTuple("Ts")


_type_name = partial(type_name, short=True)


@dataclass
class MyGenericDataClassTs(Generic[Unpack[Ts]], DataClassDictMixin):
    x: Tuple[int, Unpack[Ts]]


class MyGenericTs(Generic[Unpack[Ts]]):
    pass


class MyGenericTsK(Generic[Unpack[Ts], K]):
    pass


def test_check_generic():
    with pytest.raises(TypeError) as e:
        _check_generic(object, (K, Unpack[Ts], Unpack[Ts]), (int,))
    assert (
        str(e.value)
        == "Multiple unpacks are disallowed within a single type parameter "
        "list for object"
    )

    with pytest.raises(TypeError) as e:
        _check_generic(object, (K, V), (int,))
    assert (
        str(e.value) == f"Too few arguments for object; actual 1, expected 2"
    )

    with pytest.raises(TypeError) as e:
        _check_generic(object, (K, Unpack[Ts], V), (int,))
    assert (
        str(e.value)
        == "Too few arguments for object; actual 1, expected at least 2"
    )


def test_dataclass_with_multiple_unpacks():
    with pytest.raises(TypeError) as e:

        @dataclass
        class DataClass(DataClassDictMixin):
            x: Tuple[Unpack[Tuple[int]], Unpack[Tuple[float]]]

    typ_name = type_name(Tuple[Unpack[Tuple[int]], Unpack[Tuple[float]]])
    assert (
        str(e.value)
        == "Multiple unpacks are disallowed within a single type parameter "
        f"list for {typ_name}"
    )


def test_dataclass_with_single_unpack_tuple():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: Tuple[Unpack[Tuple[int, ...]]]
        b: Tuple[Unpack[Tuple[int, float, int]]]
        c: Tuple[Unpack[Tuple[int]]]
        d: Tuple[Unpack[Tuple[()]]]

    obj = DataClass(
        a=(1, 2, 3, 4, 5),
        b=(1, 2.2, 3),
        c=(1,),
        d=(),
    )
    assert obj.to_dict() == {
        "a": [1, 2, 3, 4, 5],
        "b": [1, 2.2, 3],
        "c": [1],
        "d": [],
    }
    assert (
        DataClass.from_dict(
            {
                "a": ["1", "2", "3", "4", 5.0],
                "b": ["1", "2.2", "3"],
                "c": ["1"],
                "d": ["1", "2", "3"],
            }
        )
        == obj
    )


def test_dataclass_with_mixed_unpack_tuple_ellipsis():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: Tuple[float, Unpack[Tuple[int, ...]]]
        b: Tuple[Unpack[Tuple[int, ...]], float]
        c: Tuple[float, Unpack[Tuple[int, ...]], float]

    obj = DataClass(
        a=(1.1, 2, 3, 4, 5),
        b=(1, 2, 3, 4, 5.5),
        c=(1.1, 2, 3, 4, 5.5),
    )
    assert obj.to_dict() == {
        "a": [1.1, 2, 3, 4, 5],
        "b": [1, 2, 3, 4, 5.5],
        "c": [1.1, 2, 3, 4, 5.5],
    }
    assert (
        DataClass.from_dict(
            {
                "a": ["1.1", "2", "3", "4", 5.0],
                "b": [1.0, "2", "3", "4", "5.5"],
                "c": ["1.1", "2", "3", 4.0, 5.5],
            }
        )
        == obj
    )


def test_dataclass_with_mixed_unpack_tuple_multiple_args():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: Tuple[float, Unpack[Tuple[int, float, int]]]
        b: Tuple[Unpack[Tuple[int, float, int]], float]
        c: Tuple[float, Unpack[Tuple[int, float, int]], float]

    obj = DataClass(
        a=(1.1, 2, 3.3, 4),
        b=(1, 2.2, 3, 4.4),
        c=(1.1, 2, 3.3, 4, 5.5),
    )
    assert obj.to_dict() == {
        "a": [1.1, 2, 3.3, 4],
        "b": [1, 2.2, 3, 4.4],
        "c": [1.1, 2, 3.3, 4, 5.5],
    }
    assert (
        DataClass.from_dict(
            {
                "a": ["1.1", "2", 3.3, 4.0],
                "b": [1.0, 2.2, "3", "4.4"],
                "c": ["1.1", "2", 3.3, 4.0, "5.5"],
            }
        )
        == obj
    )


def test_dataclass_with_mixed_unpack_tuple_one_arg():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: Tuple[float, Unpack[Tuple[int]]]
        b: Tuple[Unpack[Tuple[int]], float]
        c: Tuple[float, Unpack[Tuple[int]], float]

    obj = DataClass(
        a=(1.1, 2),
        b=(1, 2.2),
        c=(1.1, 2, 3.3),
    )
    assert obj.to_dict() == {
        "a": [1.1, 2],
        "b": [1, 2.2],
        "c": [1.1, 2, 3.3],
    }
    assert (
        DataClass.from_dict(
            {
                "a": ["1.1", 2.0],
                "b": [1.0, "2.2"],
                "c": ["1.1", 2.0, 3.3],
            }
        )
        == obj
    )


def test_dataclass_with_mixed_unpack_empty_tuple():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: Tuple[float, Unpack[Tuple[()]]]
        b: Tuple[Unpack[Tuple[()]], float]
        c: Tuple[float, Unpack[Tuple[()]], float]

    obj = DataClass(
        a=(1.1,),
        b=(1.1,),
        c=(1.1, 2.2),
    )
    assert obj.to_dict() == {
        "a": [1.1],
        "b": [1.1],
        "c": [1.1, 2.2],
    }
    assert (
        DataClass.from_dict(
            {
                "a": ["1.1"],
                "b": ["1.1"],
                "c": ["1.1", "2.2"],
            }
        )
        == obj
    )


@pytest.mark.skipif(not PY_311_MIN, reason="requires python 3.11")
def test_type_name_for_unpacks_py_311():
    assert _type_name(Tuple[Unpack[Tuple[int, ...]]]) == "Tuple[int, ...]"
    assert _type_name(Tuple[Unpack[Tuple[int, float]]]) == "Tuple[int, float]"
    assert (
        _type_name(Tuple[int, Unpack[Tuple[float, ...]]])
        == "Tuple[int, *Tuple[float, ...]]"
    )
    assert (
        _type_name(Tuple[int, Unpack[Tuple[float, str]]])
        == "Tuple[int, float, str]"
    )
    assert _type_name(Tuple[Unpack[Tuple[()]]]) == "Tuple[()]"
    assert _type_name(Tuple[int, Unpack[Tuple[()]]]) == "Tuple[int]"
    assert (
        _type_name(Tuple[str, Unpack[Tuple[int, ...]], int])
        == "Tuple[str, *Tuple[int, ...], int]"
    )
    assert (
        _type_name(Tuple[str, Unpack[Tuple[Tuple[int], ...]], int])
        == "Tuple[str, *Tuple[Tuple[int], ...], int]"
    )
    assert (
        _type_name(
            Tuple[str, Unpack[Tuple[Tuple[Unpack[Tuple[int]], ...]]], int]
        )
        == "Tuple[str, Tuple[int, ...], int]"
    )
    assert _type_name(Tuple[Unpack[Ts]]) == "Tuple[*Ts]"
    assert _type_name(Tuple[int, Unpack[Ts], int]) == "Tuple[int, *Ts, int]"
    assert _type_name(Generic[Unpack[Ts]]) == "Generic[*Ts]"
    assert _type_name(Generic[K, Unpack[Ts], V]) == "Generic[Any, *Ts, Any]"
    assert _type_name(Unpack[Tuple[int]]) == "int"
    assert _type_name(Unpack[Tuple[int, float]]) == "int, float"
    assert _type_name(Unpack[Ts]) == "*Ts"
    assert _type_name(Ts) == "Ts"
    assert (
        _type_name(Tuple[Unpack[Ts], K][Unpack[Tuple[int, ...]]])
        == "Tuple[*Tuple[int, ...], int]"
    )
    assert (
        _type_name(Tuple[Unpack[Ts], K][Unpack[Tuple[int, ...]], date])
        == "Tuple[*Tuple[int, ...], date]"
    )
    assert (
        _type_name(Tuple[Unpack[Ts], K][date, Unpack[Tuple[int, ...]]])
        == "Tuple[date, *Tuple[int, ...], int]"
    )


@pytest.mark.skipif(PY_311_MIN, reason="requires python<3.11")
def test_type_name_for_unpacks_py_less_than_311():
    assert _type_name(Tuple[Unpack[Tuple[int, ...]]]) == "Tuple[int, ...]"
    assert _type_name(Tuple[Unpack[Tuple[int, float]]]) == "Tuple[int, float]"
    assert (
        _type_name(Tuple[int, Unpack[Tuple[float, ...]]])
        == "Tuple[int, Unpack[Tuple[float, ...]]]"
    )
    assert (
        _type_name(Tuple[int, Unpack[Tuple[float, str]]])
        == "Tuple[int, float, str]"
    )
    assert _type_name(Tuple[Unpack[Tuple[()]]]) == "Tuple[()]"
    assert _type_name(Tuple[int, Unpack[Tuple[()]]]) == "Tuple[int]"
    assert (
        _type_name(Tuple[str, Unpack[Tuple[int, ...]], int])
        == "Tuple[str, Unpack[Tuple[int, ...]], int]"
    )
    assert (
        _type_name(Tuple[str, Unpack[Tuple[Tuple[int], ...]], int])
        == "Tuple[str, Unpack[Tuple[Tuple[int], ...]], int]"
    )
    assert (
        _type_name(
            Tuple[str, Unpack[Tuple[Tuple[Unpack[Tuple[int]], ...]]], int]
        )
        == "Tuple[str, Tuple[int, ...], int]"
    )
    assert _type_name(Tuple[Unpack[Ts]]) == "Tuple[Unpack[Ts]]"
    assert (
        _type_name(Tuple[int, Unpack[Ts], int])
        == "Tuple[int, Unpack[Ts], int]"
    )
    assert _type_name(Generic[Unpack[Ts]]) == "Generic[Unpack[Ts]]"
    assert (
        _type_name(Generic[K, Unpack[Ts], V])
        == "Generic[Any, Unpack[Ts], Any]"
    )
    assert _type_name(Unpack[Tuple[int]]) == "int"
    assert _type_name(Unpack[Tuple[int, float]]) == "int, float"
    assert _type_name(Unpack[Ts]) == "Unpack[Ts]"
    assert _type_name(Ts) == "Ts"

    # this doesn't work on python<3.11
    # assert (
    #     _type_name(Tuple[Unpack[Ts], K][Unpack[Tuple[int, ...]]])
    #     == "Tuple[Unpack[Tuple[int, ...]], int]"
    # )
    # assert (
    #     _type_name(Tuple[Unpack[Ts], K][Unpack[Tuple[int, ...]], date])
    #     == "Tuple[Unpack[Tuple[int, ...]], date]"
    # )
    # assert (
    #     _type_name(Tuple[Unpack[Ts], K][date, Unpack[Tuple[int, ...]]])
    #     == "Tuple[date, Unpack[Tuple[int, ...]], int]"
    # )


def test_concrete_generic_with_empty_tuple():
    @dataclass
    # I tried MyGenericDataClassTs[()] but [()] becomes just (). Impossible.
    class ConcreteDataClass(MyGenericDataClassTs[Unpack[Tuple[()]]]):
        pass

    obj = ConcreteDataClass((1,))
    assert obj.to_dict() == {"x": [1]}
    assert ConcreteDataClass.from_dict({"x": [1]}) == obj
    assert ConcreteDataClass.from_dict({"x": [1, 2, 3]}) == obj


def test_concrete_generic_with_variable_length_tuple_any():
    @dataclass
    class ConcreteDataClass(MyGenericDataClassTs):
        pass

    obj = ConcreteDataClass((1, "a", date(2022, 12, 17)))
    assert obj.to_dict() == {"x": [1, "a", date(2022, 12, 17)]}
    assert (
        ConcreteDataClass.from_dict({"x": ["1", "a", date(2022, 12, 17)]})
        == obj
    )


def test_concrete_generic_with_replaced_tuple_with_one_arg():
    @dataclass
    class ConcreteDataClass(MyGenericDataClassTs[Unpack[Tuple[K]]]):
        pass

    obj = ConcreteDataClass((1, date(2022, 12, 17)))
    assert obj.to_dict() == {"x": [1, date(2022, 12, 17)]}
    assert (
        ConcreteDataClass.from_dict({"x": ["1", date(2022, 12, 17), "a"]})
        == obj
    )


def test_concrete_generic_with_replaced_tuple_with_multiple_args():
    @dataclass
    class ConcreteDataClass(MyGenericDataClassTs[Unpack[Tuple[float, float]]]):
        pass

    obj = ConcreteDataClass((1, 2.2, 3.3))
    assert obj.to_dict() == {"x": [1, 2.2, 3.3]}
    assert ConcreteDataClass.from_dict({"x": ["1", "2.2", "3.3"]}) == obj


@pytest.mark.skipif(not PY_311_MIN, reason="doesn't work on py<3.11")
def test_with_int_float_tuple_and_any_at_the_end():
    Ts1 = TypeVarTuple("Ts1")
    Ts2 = TypeVarTuple("Ts2")
    IntTuple = Tuple[int, Unpack[Ts1]]
    IntFloatTuple = IntTuple[float, Unpack[Ts2]]

    @dataclass
    class DataClass(DataClassDictMixin):
        x: IntFloatTuple

    obj = DataClass((1, 2.2, "3", date(2022, 12, 17)))
    assert obj.to_dict() == {"x": [1, 2.2, "3", date(2022, 12, 17)]}
    assert (
        DataClass.from_dict({"x": ["1", "2.2", "3", date(2022, 12, 17)]})
        == obj
    )


@pytest.mark.skipif(not PY_311_MIN, reason="doesn't work on py<3.11")
def test_with_int_floats_tuple():
    Ts1 = TypeVarTuple("Ts1")
    IntTuple = Tuple[int, Unpack[Ts1]]
    IntFloatsTuple = IntTuple[Unpack[Tuple[float, ...]]]

    @dataclass
    class DataClass(DataClassDictMixin):
        x: IntFloatsTuple

    obj = DataClass((1, 2.2, 3.3, 4.4))
    assert obj.to_dict() == {"x": [1, 2.2, 3.3, 4.4]}
    assert DataClass.from_dict({"x": ["1", "2.2", "3.3", "4.4"]}) == obj


@pytest.mark.skipif(not PY_311_MIN, reason="doesn't work on py<3.11")
def test_splitting_arbitrary_length_tuples_1():
    Elderberries = Tuple[Unpack[Ts], K]

    @dataclass
    class DataClass(DataClassDictMixin):
        x: Elderberries[Unpack[Tuple[int, ...]]]

    obj = DataClass((1, 2, 3))
    assert obj.to_dict() == {"x": [1, 2, 3]}
    assert DataClass.from_dict({"x": ["1", "2", "3"]}) == obj

    obj = DataClass((1,))
    assert obj.to_dict() == {"x": [1]}
    assert DataClass.from_dict({"x": [1]}) == obj


def test_dataclass_with_splitting_arbitrary_length_tuples_1():
    @dataclass
    class GenericDataClass(Generic[Unpack[Ts], K], DataClassDictMixin):
        x: Tuple[Unpack[Ts], K]

    @dataclass
    class ConcreteDataClass(GenericDataClass[Unpack[Tuple[int, ...]]]):
        pass

    obj = ConcreteDataClass((1, 2, 3))
    assert obj.to_dict() == {"x": [1, 2, 3]}
    assert ConcreteDataClass.from_dict({"x": ["1", "2", "3"]}) == obj

    obj = ConcreteDataClass((1,))
    assert obj.to_dict() == {"x": [1]}
    assert ConcreteDataClass.from_dict({"x": [1]}) == obj


@pytest.mark.skipif(not PY_311_MIN, reason="doesn't work on py<3.11")
def test_splitting_arbitrary_length_tuples_2():
    Elderberries = Tuple[Unpack[Ts], K]

    @dataclass
    class DataClass(DataClassDictMixin):
        x: Elderberries[Unpack[Tuple[int, ...]], date]

    obj = DataClass((1, 2, date(2022, 12, 17)))
    assert obj.to_dict() == {"x": [1, 2, "2022-12-17"]}
    assert DataClass.from_dict({"x": ["1", "2", "2022-12-17"]}) == obj

    obj = DataClass((date(2022, 12, 17),))
    assert obj.to_dict() == {"x": ["2022-12-17"]}
    assert DataClass.from_dict({"x": ["2022-12-17"]}) == obj


def test_dataclass_with_splitting_arbitrary_length_tuples_2():
    @dataclass
    class GenericDataClass(Generic[Unpack[Ts], K], DataClassDictMixin):
        x: Tuple[Unpack[Ts], K]

    @dataclass
    class ConcreteDataClass(GenericDataClass[Unpack[Tuple[int, ...]], date]):
        pass

    obj = ConcreteDataClass((1, 2, date(2022, 12, 17)))
    assert obj.to_dict() == {"x": [1, 2, "2022-12-17"]}
    assert ConcreteDataClass.from_dict({"x": ["1", "2", "2022-12-17"]}) == obj

    obj = ConcreteDataClass((date(2022, 12, 17),))
    assert obj.to_dict() == {"x": ["2022-12-17"]}
    assert ConcreteDataClass.from_dict({"x": ["2022-12-17"]}) == obj


@pytest.mark.skipif(not PY_311_MIN, reason="doesn't work on py<3.11")
def test_splitting_arbitrary_length_tuples_3():
    Elderberries = Tuple[Unpack[Ts], K]

    @dataclass
    class DataClass(DataClassDictMixin):
        x: Elderberries[date, Unpack[Tuple[int, ...]]]

    obj = DataClass((date(2022, 12, 17), 1, 2, 3))
    assert obj.to_dict() == {"x": ["2022-12-17", 1, 2, 3]}
    assert DataClass.from_dict({"x": ["2022-12-17", "1", "2", "3"]}) == obj

    obj = DataClass((date(2022, 12, 17), 1))
    assert obj.to_dict() == {"x": ["2022-12-17", 1]}
    assert DataClass.from_dict({"x": ["2022-12-17", "1"]}) == obj


def test_dataclass_with_splitting_arbitrary_length_tuples_3():
    @dataclass
    class GenericDataClass(Generic[Unpack[Ts], K], DataClassDictMixin):
        x: Tuple[Unpack[Ts], K]

    @dataclass
    class ConcreteDataClass(GenericDataClass[date, Unpack[Tuple[int, ...]]]):
        pass

        class Config:
            debug = True

    obj = ConcreteDataClass((date(2022, 12, 17), 1, 2))
    assert obj.to_dict() == {"x": ["2022-12-17", 1, 2]}
    assert ConcreteDataClass.from_dict({"x": ["2022-12-17", "1", "2"]}) == obj

    obj = ConcreteDataClass((date(2022, 12, 17), 1))
    assert obj.to_dict() == {"x": ["2022-12-17", 1]}
    assert ConcreteDataClass.from_dict({"x": ["2022-12-17", 1]}) == obj


def test_resolve_type_params_with_unpacks():
    assert resolve_type_params(MyGenericTsK, [int, float], False) == {
        MyGenericTsK: {K: float, Unpack[Ts]: Unpack[Tuple[int]]}
    }
    assert resolve_type_params(MyGenericTsK, [int, str, float], False) == {
        MyGenericTsK: {K: float, Unpack[Ts]: Unpack[Tuple[int, str]]}
    }
    assert resolve_type_params(
        MyGenericTsK, [Unpack[Tuple[int, str]], float], False
    ) == {MyGenericTsK: {K: float, Unpack[Ts]: Unpack[Tuple[int, str]]}}
    assert resolve_type_params(
        MyGenericTsK, [Unpack[Tuple[int, ...]]], False
    ) == {MyGenericTsK: {K: int, Unpack[Ts]: Unpack[Tuple[int, ...]]}}
    assert resolve_type_params(
        MyGenericTsK, [str, Unpack[Tuple[int, ...]]], False
    ) == {
        MyGenericTsK: {
            K: int,
            Unpack[Ts]: Unpack[Tuple[str, Unpack[Tuple[int, ...]]]],
        }
    }
    assert resolve_type_params(MyGenericTs, [()], False) == {
        MyGenericTs: {Unpack[Ts]: Unpack[Tuple[()]]}
    }


def test_dataclass_with_tuple_int_and_empty():
    @dataclass
    class ConcreteDataClass(MyGenericDataClassTs[Unpack[Tuple[()]]]):
        pass

    obj = ConcreteDataClass((1,))
    assert obj.to_dict() == {"x": [1]}
    assert ConcreteDataClass.from_dict({"x": [1]}) == obj
    assert ConcreteDataClass.from_dict({"x": [1, 2, 3]}) == obj
    with pytest.raises(MissingField):
        ConcreteDataClass.from_dict({})


def test_unpack_tuple_with_multiple_unpacks():
    spec = ValueSpec(
        type=Tuple,
        expression="value",
        builder=object,
        field_ctx=FieldContext("x", {}),
    )
    with pytest.raises(TypeError):
        unpack_tuple(spec, (Unpack[Tuple[int]], Unpack[Tuple[float]]))
    with pytest.raises(TypeError):
        pack_tuple(spec, (Unpack[Tuple[int]], Unpack[Tuple[float]]))


def test_flatten_type_args_with_empty_tuple():
    assert _flatten_type_args([Unpack[Tuple[()]]]) == [()]
    assert _flatten_type_args([int, Unpack[Tuple[()]]]) == [int]
    assert _flatten_type_args([Unpack[tuple[()]]]) == [()]
    assert _flatten_type_args([int, Unpack[tuple[()]]]) == [int]



import pytest

from mashumaro.config import BaseConfig
from mashumaro.jsonschema.models import (
    JSONSchemaStringFormat,
    _deserialize_json_schema_instance_format,
)
from mashumaro.jsonschema.schema import Instance


def test_instance_get_configs():
    instance = Instance(int)
    assert instance.get_owner_config() is BaseConfig
    assert instance.get_self_config() is BaseConfig

    derived = instance.derive()
    assert derived.get_self_config() is instance.get_self_config()


def test_deserialize_json_schema_instance_format():
    assert (
        _deserialize_json_schema_instance_format("email")
        is JSONSchemaStringFormat.EMAIL
    )
    with pytest.raises(ValueError):
        assert _deserialize_json_schema_instance_format("foobar")



from __future__ import annotations

from dataclasses import dataclass
from datetime import date
from typing import Dict, Generic, TypeVar

from mashumaro import DataClassDictMixin
from mashumaro.config import BaseConfig
from mashumaro.types import SerializationStrategy

T = TypeVar("T")


class Foo(Generic[T]):
    a: T

    def __init__(self, a: T) -> None:
        self.a = a

    def __eq__(self, other: Foo) -> bool:
        return self.a == other.a


class FooStrategy(Generic[T], SerializationStrategy):
    def deserialize(self, value: Dict[str, T]) -> Foo[T]:
        return Foo(**value)

    def serialize(self, value: Foo[T]) -> Dict[str, T]:
        return {"a": value.a}


@dataclass
class Bar(DataClassDictMixin):
    x_str: Foo[str]
    x_date: Foo[date]

    class Config(BaseConfig):
        serialization_strategy = {Foo: FooStrategy()}


def test_generic_serialization_strategy():
    data = {"x_str": {"a": "2023-11-14"}, "x_date": {"a": "2023-11-14"}}
    obj = Bar(Foo("2023-11-14"), Foo(date(2023, 11, 14)))
    assert obj.to_dict() == data
    assert Bar.from_dict(data) == obj



import dataclasses
from dataclasses import dataclass, field
from enum import Enum
from pathlib import PurePosixPath
from typing import Any, Literal, NamedTuple, Optional

import msgpack
import pytest
from typing_extensions import Self

from mashumaro.config import BaseConfig
from mashumaro.core.const import PY_310_MIN
from mashumaro.core.meta.types.common import clean_id
from mashumaro.mixins.dict import DataClassDictMixin
from mashumaro.mixins.json import DataClassJSONMixin
from mashumaro.mixins.msgpack import DataClassMessagePackMixin
from mashumaro.mixins.yaml import DataClassYAMLMixin
from mashumaro.types import GenericSerializableType, SerializableType


@dataclass
class EntityA1(DataClassDictMixin):
    x: int


@dataclass
class EntityA2(EntityA1):
    y: int


@dataclass
class EntityA1Wrapper(DataClassMessagePackMixin):
    entity: EntityA1


@dataclass
class EntityA2Wrapper(DataClassMessagePackMixin):
    entity: EntityA2


@dataclass
class EntityB1(DataClassDictMixin):
    x: int


@dataclass
class EntityB2(EntityB1):
    y: int


@dataclass
class EntityB1WrapperDict(DataClassDictMixin):
    entity: EntityB1


@dataclass
class EntityB2WrapperMessagePack(DataClassMessagePackMixin):
    entity: EntityB2


@dataclass
class EntityBWrapperMessagePack(DataClassMessagePackMixin):
    entity1wrapper: EntityB1WrapperDict
    entity2wrapper: EntityB2WrapperMessagePack


if PY_310_MIN:

    @dataclass(kw_only=True)
    class DataClassKwOnly1(DataClassDictMixin):
        x: int
        y: int

    @dataclass
    class DataClassKwOnly2(DataClassDictMixin):
        x: int = field(kw_only=True)
        y: int

    @dataclass(kw_only=True)
    class DataClassKwOnly3(DataClassDictMixin):
        x: int
        y: int = field(kw_only=False)

    @dataclass
    class DataClassKwOnly4(DataClassDictMixin):
        x: int
        _: dataclasses.KW_ONLY
        y: int

    @dataclass(kw_only=True)
    class LazyDataClassKwOnly1(DataClassDictMixin):
        x: int
        y: int

        class Config(BaseConfig):
            lazy_compilation = True

    @dataclass
    class LazyDataClassKwOnly2(DataClassDictMixin):
        x: int = field(kw_only=True)
        y: int

        class Config(BaseConfig):
            lazy_compilation = True

    @dataclass(kw_only=True)
    class LazyDataClassKwOnly3(DataClassDictMixin):
        x: int
        y: int = field(kw_only=False)

        class Config(BaseConfig):
            lazy_compilation = True

    @dataclass
    class LazyDataClassKwOnly4(DataClassDictMixin):
        x: int
        _: dataclasses.KW_ONLY
        y: int

        class Config(BaseConfig):
            lazy_compilation = True


@dataclass
class BaseClassWithPosArgs(DataClassDictMixin):
    pos1: int
    pos2: int
    pos3: int


@dataclass
class DerivedClassWithOverriddenMiddlePosArg(BaseClassWithPosArgs):
    kw1: int = 0
    pos2: int


@dataclass
class DerivedClassWithOverriddenMiddlePosArgWithField(BaseClassWithPosArgs):
    kw1: int = 0
    pos2: int = field(metadata={})


def test_slots():
    @dataclass
    class RegularDataClass:
        __slots__ = ("number",)
        number: int

    @dataclass
    class DictDataClass(DataClassDictMixin):
        __slots__ = ("number",)
        number: int

    @dataclass
    class JSONDataClass(DataClassJSONMixin):
        __slots__ = ("number",)
        number: int

    @dataclass
    class MessagePackDataClass(DataClassMessagePackMixin):
        __slots__ = ("number",)
        number: int

    @dataclass
    class YAMLDataClass(DataClassYAMLMixin):
        __slots__ = ("number",)
        number: int

    class MySerializableType(SerializableType):
        __slots__ = ("number",)

        def __init__(self, number):
            self.number = number

    class MyGenericSerializableType(GenericSerializableType):
        __slots__ = ("number",)

        def __init__(self, number):
            self.number = number

    for cls in (
        RegularDataClass,
        DictDataClass,
        JSONDataClass,
        MessagePackDataClass,
        YAMLDataClass,
        MySerializableType,
        MyGenericSerializableType,
    ):
        instance = cls(1)
        with pytest.raises(AttributeError) as e:
            instance.new_attribute = 2
        assert str(e.value).startswith(
            f"'{cls.__name__}' object has no attribute 'new_attribute'"
        )


def test_data_class_dict_mixin_from_dict():
    assert DataClassDictMixin.from_dict({}) is None


def test_data_class_dict_mixin_to_dict():
    assert DataClassDictMixin().to_dict() is None


def test_compiled_mixin_with_inheritance_1():
    entity = EntityA2(x=1, y=2)
    wrapper = EntityA2Wrapper(entity)
    data = msgpack.packb({"entity": {"x": 1, "y": 2}}, use_bin_type=True)
    assert wrapper.to_msgpack() == data
    assert EntityA2Wrapper.from_msgpack(data) == wrapper


def test_compiled_mixin_with_inheritance_2():
    entity1w = EntityB1WrapperDict(EntityB1(x=1))
    entity2w = EntityB2WrapperMessagePack(EntityB2(x=1, y=2))
    wrapper = EntityBWrapperMessagePack(entity1w, entity2w)
    data = msgpack.packb(
        {
            "entity1wrapper": {"entity": {"x": 1}},
            "entity2wrapper": {"entity": {"x": 1, "y": 2}},
        },
        use_bin_type=True,
    )
    assert wrapper.to_msgpack() == data
    assert EntityBWrapperMessagePack.from_msgpack(data) == wrapper


@pytest.mark.skipif(not PY_310_MIN, reason="requires python 3.10+")
def test_kw_only_dataclasses():
    data = {"x": "1", "y": "2"}
    for cls in (
        DataClassKwOnly1,
        DataClassKwOnly2,
        DataClassKwOnly3,
        DataClassKwOnly4,
        LazyDataClassKwOnly1,
        LazyDataClassKwOnly2,
        LazyDataClassKwOnly3,
        LazyDataClassKwOnly4,
    ):
        obj = cls.from_dict(data)
        assert obj.x == 1
        assert obj.y == 2


def test_kw_args_when_pos_arg_is_overridden_without_field():
    obj = DerivedClassWithOverriddenMiddlePosArg(1, 2, 3, 4)
    loaded = DerivedClassWithOverriddenMiddlePosArg.from_dict(
        {"pos1": "1", "pos2": "2", "pos3": "3", "kw1": "4"}
    )
    assert loaded == obj
    assert loaded.pos1 == 1
    assert loaded.pos2 == 2
    assert loaded.pos3 == 3
    assert loaded.kw1 == 4


def test_kw_args_when_pos_arg_is_overridden_with_field():
    obj = DerivedClassWithOverriddenMiddlePosArgWithField(1, 2, 3, 4)
    loaded = DerivedClassWithOverriddenMiddlePosArgWithField.from_dict(
        {"pos1": "1", "pos2": "2", "pos3": "3", "kw1": "4"}
    )
    assert loaded == obj
    assert loaded.pos1 == 1
    assert loaded.pos2 == 2
    assert loaded.pos3 == 3
    assert loaded.kw1 == 4


def test_local_types():
    @dataclass
    class LocalDataclassType:
        foo: int

    class LocalNamedTupleType(NamedTuple):
        foo: int

    class LocalPathLike(PurePosixPath):
        pass

    class LocalEnumType(Enum):
        FOO = "foo"

    class LocalSerializableType(SerializableType):
        @classmethod
        def _deserialize(cls, value):
            return LocalSerializableType()

        def _serialize(self) -> Any:
            return {}

        def __eq__(self, __value: object) -> bool:
            return isinstance(__value, LocalSerializableType)

    class LocalGenericSerializableType(GenericSerializableType):
        @classmethod
        def _deserialize(cls, value, types):
            return LocalGenericSerializableType()

        def _serialize(self, types) -> Any:
            return {}

        def __eq__(self, __value: object) -> bool:
            return isinstance(__value, LocalGenericSerializableType)

    class LocalSelfSerializableAnnotatedType(
        SerializableType, use_annotations=True
    ):
        @classmethod
        def _deserialize(cls, value: Self) -> Self:
            return value

        def _serialize(self) -> Self:
            return self

        def __eq__(self, __value: object) -> bool:
            return isinstance(__value, LocalSelfSerializableAnnotatedType)

    @dataclass
    class DataClassWithLocalType(DataClassDictMixin):
        x1: LocalDataclassType
        x2: LocalNamedTupleType
        x3: LocalPathLike
        x4: LocalEnumType
        x4_1: Literal[LocalEnumType.FOO]
        x5: LocalSerializableType
        x6: LocalGenericSerializableType
        x7: Optional[Self]
        x8: LocalSelfSerializableAnnotatedType

    obj = DataClassWithLocalType(
        x1=LocalDataclassType(foo=0),
        x2=LocalNamedTupleType(foo=0),
        x3=LocalPathLike("path/to/file"),
        x4=LocalEnumType.FOO,
        x4_1=LocalEnumType.FOO,
        x5=LocalSerializableType(),
        x6=LocalGenericSerializableType(),
        x7=None,
        x8=LocalSelfSerializableAnnotatedType(),
    )
    assert obj.to_dict() == {
        "x1": {"foo": 0},
        "x2": [0],
        "x3": "path/to/file",
        "x4": "foo",
        "x4_1": "foo",
        "x5": {},
        "x6": {},
        "x7": None,
        "x8": LocalSelfSerializableAnnotatedType(),
    }
    assert (
        DataClassWithLocalType.from_dict(
            {
                "x1": {"foo": 0},
                "x2": [0],
                "x3": "path/to/file",
                "x4": "foo",
                "x4_1": "foo",
                "x5": {},
                "x6": {},
                "x7": None,
                "x8": LocalSelfSerializableAnnotatedType(),
            }
        )
        == obj
    )


def test_clean_id():
    assert clean_id("") == "_"
    assert clean_id("foo") == "foo"
    assert clean_id("foo.<locals>.bar") == "foo__locals__bar"



import json
from pathlib import Path

from pyperf import Benchmark


def create_spec(benchmark_type: str) -> None:
    data_dir = Path(Path.cwd() / "benchmark" / "data")
    with open(data_dir / "spec_template.json") as f:
        spec = json.load(f)
    if benchmark_type == "load":
        spec["title"]["text"] = "Creating GitHub Issue object from dict"
    elif benchmark_type == "dump":
        spec["title"]["text"] = "Converting GitHub Issue object to dict"
    values = spec["data"]["values"]

    for file in Path(Path.cwd() / "benchmark" / "data" / "results").glob(
        f"{benchmark_type}_*.json"
    ):
        benchmark: Benchmark = Benchmark.load(str(file))
        library_name = benchmark.get_name()[:-6]
        values.append(
            {
                "library": library_name,
                "time": benchmark.mean(),
                "timeFormat": benchmark.format_value(benchmark.mean()),
            }
        )
    values.sort(key=lambda v: v["time"])

    with open(data_dir / f"spec_{benchmark_type}.json", "w") as f:
        json.dump(spec, f)


if __name__ == "__main__":
    create_spec("load")
    create_spec("dump")



import pytest

from mashumaro.types import Discriminator


def test_discriminator_without_necessary_params():
    with pytest.raises(ValueError) as exc_info:
        Discriminator()
    assert str(exc_info.value) == (
        "Either 'include_supertypes' or 'include_subtypes' must be enabled"
    )

    with pytest.raises(ValueError) as exc_info:
        Discriminator(field="type")
    assert str(exc_info.value) == (
        "Either 'include_supertypes' or 'include_subtypes' must be enabled"
    )



from dataclasses import dataclass
from datetime import date
from typing import Any, Generic, Iterator, Mapping, Tuple, TypeVar

import pytest
from typing_extensions import Self

from mashumaro import DataClassDictMixin
from mashumaro.exceptions import UnserializableField
from mashumaro.types import Alias, SerializableType
from tests.entities import GenericSerializableList, GenericSerializableWrapper

XT = TypeVar("XT")
YT = TypeVar("YT")


class MySerializableType(SerializableType):
    def __init__(self, value):
        self.value = value

    def _serialize(self):
        return self.value.isoformat()

    @classmethod
    def _deserialize(cls, value):
        return cls(date.fromisoformat(value))


class MyAnnotatedSerializableType(SerializableType, use_annotations=True):
    def __init__(self, value: date):
        self.value = value

    def _serialize(self) -> date:
        return self.value

    @classmethod
    def _deserialize(cls, value: date):
        return cls(value)


class MyAnnotatedSerializableTypeWithoutAnnotationsInDeserialize(
    SerializableType, use_annotations=True
):
    def __init__(self, value):  # pragma: no cover
        self.value = value

    def _serialize(self) -> date:  # pragma: no cover
        return self.value

    @classmethod
    def _deserialize(cls, value):  # pragma: no cover
        return cls(value)


class MyAnnotatedSerializableTypeWithoutAnnotationsInSerialize(
    SerializableType, use_annotations=True
):
    def __init__(self, value):  # pragma: no cover
        self.value = value

    def _serialize(self):  # pragma: no cover
        return self.value

    @classmethod
    def _deserialize(cls, value: date):  # pragma: no cover
        return cls(value)


class MyAnnotatedGenericSerializableType(
    Generic[XT, YT], SerializableType, use_annotations=True
):
    def __init__(self, value: Tuple[XT, YT]):
        self.value = value

    def _serialize(self) -> Tuple[YT, XT]:
        return tuple(reversed(self.value))

    @classmethod
    def _deserialize(cls, value: Tuple[XT, YT]):
        return cls(value)


class MyAnnotatedGenericSerializableTypeWithMixedTypeVars(
    Generic[XT], SerializableType, use_annotations=True
):
    def __init__(self, value):
        self.value = value

    def _serialize(self) -> Tuple[int, XT, float]:
        return self.value

    @classmethod
    def _deserialize(cls, value: Tuple[int, XT, float]):
        return cls(value)


class MyAnnotatedGenericPEP585SerializableType(
    Generic[XT, YT], SerializableType, use_annotations=True
):
    def __init__(self, value: tuple[XT, YT]):
        self.value = value

    def _serialize(self) -> tuple[YT, XT]:
        return tuple(reversed(self.value))

    @classmethod
    def _deserialize(cls, value: tuple[XT, YT]):
        return cls(value)


class MyMapping(Mapping[XT, YT], SerializableType, use_annotations=True):
    def __init__(self, value: Mapping[Any, Any]):
        self.value = value

    def _serialize(self) -> Mapping[XT, YT]:
        return self.value

    @classmethod
    def _deserialize(cls, mapping: Mapping[XT, YT]):
        return cls(mapping)

    def __getitem__(self, __k: XT) -> YT:
        return self.value[__k]

    def __len__(self) -> int:  # pragma: no cover
        return len(self.value)

    def __iter__(self) -> Iterator[XT]:
        return iter(self.value)

    def __repr__(self):  # pragma: no cover
        return f"<MyMapping: {repr(self.value)}>"


class MyAnnotatedUserGenericSerializableType(
    Generic[XT, YT], SerializableType, use_annotations=True
):
    def __init__(self, value: MyMapping[XT, YT]):
        self.value = value

    def _serialize(self) -> MyMapping[XT, YT]:
        return self.value

    @classmethod
    def _deserialize(cls, value: MyMapping[XT, YT]):
        return cls(value)

    def __repr__(self):  # pragma: no cover
        return f"<MyAnnotatedUserGenericSerializableType: {repr(self.value)}>"


class MySelfSerializableType(SerializableType, use_annotations=True):
    def _serialize(self) -> Self:
        return self

    @classmethod
    def _deserialize(cls, value: Self) -> Self:
        return value


def test_generic_serializable_list_int():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: GenericSerializableList[int]

    obj = DataClass(x=GenericSerializableList([1, 2, 3]))
    assert DataClass.from_dict({"x": [3, 4, 5]}) == obj
    assert obj.to_dict() == {"x": [3, 4, 5]}


def test_generic_serializable_list_str():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: GenericSerializableList[str]

    obj = DataClass(x=GenericSerializableList(["a", "b", "c"]))
    assert DataClass.from_dict({"x": ["_a", "_b", "_c"]}) == obj
    assert obj.to_dict() == {"x": ["_a", "_b", "_c"]}


def test_generic_serializable_wrapper_with_type_from_another_module():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: GenericSerializableWrapper[date]

    obj = DataClass(x=GenericSerializableWrapper(date(2022, 12, 8)))
    assert DataClass.from_dict({"x": "2022-12-08"}) == obj
    assert obj.to_dict() == {"x": "2022-12-08"}


def test_simple_serializable_type():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MySerializableType

    obj = DataClass.from_dict({"x": "2022-12-07"})
    assert obj.x.value == date(2022, 12, 7)
    assert obj.to_dict() == {"x": "2022-12-07"}


def test_annotated_serializable_type():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyAnnotatedSerializableType

    obj = DataClass.from_dict({"x": "2022-12-07"})
    assert obj.x.value == date(2022, 12, 7)
    assert obj.to_dict() == {"x": "2022-12-07"}


def test_annotated_serializable_type_without_annotations_in_deserialize():
    with pytest.raises(UnserializableField) as e:

        @dataclass
        class _(DataClassDictMixin):
            x: MyAnnotatedSerializableTypeWithoutAnnotationsInDeserialize

    assert (
        e.value.msg
        == 'Method _deserialize must have annotated "value" argument'
    )


def test_annotated_serializable_type_without_annotations_in_serialize():
    with pytest.raises(UnserializableField) as e:

        @dataclass
        class _(DataClassDictMixin):
            x: MyAnnotatedSerializableTypeWithoutAnnotationsInSerialize

    assert e.value.msg == "Method _serialize must have return annotation"


def test_annotated_generic_serializable_type():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyAnnotatedGenericSerializableType

    obj = DataClass.from_dict({"x": ["2022-12-07", "3.14"]})
    assert obj.x.value == ("2022-12-07", "3.14")
    assert obj.to_dict() == {"x": ["3.14", "2022-12-07"]}

    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyAnnotatedGenericSerializableType[date, float]

    obj = DataClass.from_dict({"x": ["2022-12-07", "3.14"]})
    assert obj.x.value == (date(2022, 12, 7), 3.14)
    assert obj.to_dict() == {"x": [3.14, "2022-12-07"]}

    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyAnnotatedGenericSerializableType[date, YT]

    obj = DataClass.from_dict({"x": ["2022-12-07", "3.14"]})
    assert obj.x.value == (date(2022, 12, 7), "3.14")
    assert obj.to_dict() == {"x": ["3.14", "2022-12-07"]}


def test_annotated_generic_pep585_serializable_type():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyAnnotatedGenericPEP585SerializableType

    obj = DataClass.from_dict({"x": ["2022-12-07", "3.14"]})
    assert obj.x.value == ("2022-12-07", "3.14")
    assert obj.to_dict() == {"x": ["3.14", "2022-12-07"]}

    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyAnnotatedGenericSerializableType[date, float]

    obj = DataClass.from_dict({"x": ["2022-12-07", "3.14"]})
    assert obj.x.value == (date(2022, 12, 7), 3.14)
    assert obj.to_dict() == {"x": [3.14, "2022-12-07"]}

    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyAnnotatedGenericSerializableType[date, YT]

    obj = DataClass.from_dict({"x": ["2022-12-07", "3.14"]})
    assert obj.x.value == (date(2022, 12, 7), "3.14")
    assert obj.to_dict() == {"x": ["3.14", "2022-12-07"]}


def test_annotated_user_generic_serializable_type():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyAnnotatedUserGenericSerializableType

    obj = DataClass.from_dict({"x": {"2022-12-07": "3.14"}})
    assert type(obj.x.value) is MyMapping
    assert dict(obj.x.value) == {"2022-12-07": "3.14"}
    assert obj.to_dict() == {"x": {"2022-12-07": "3.14"}}

    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyAnnotatedUserGenericSerializableType[date, float]

    obj = DataClass.from_dict({"x": {"2022-12-07": "3.14"}})
    assert type(obj.x.value) is MyMapping
    assert dict(obj.x.value) == {date(2022, 12, 7): 3.14}
    assert obj.to_dict() == {"x": {"2022-12-07": 3.14}}

    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyAnnotatedUserGenericSerializableType[date, YT]

    obj = DataClass.from_dict({"x": {"2022-12-07": "3.14"}})
    assert type(obj.x.value) is MyMapping
    assert dict(obj.x.value) == {date(2022, 12, 7): "3.14"}
    assert obj.to_dict() == {"x": {"2022-12-07": "3.14"}}


def test_serializable_type_inheritance():
    class MySerializableType2(MySerializableType, use_annotations=True):
        pass

    assert MySerializableType2.__use_annotations__

    class MySerializableType3(MySerializableType2):
        pass

    assert MySerializableType3.__use_annotations__

    class MySerializableType4(MySerializableType3, use_annotations=False):
        pass

    assert not MySerializableType4.__use_annotations__


def test_serializable_type_with_self():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MySelfSerializableType

    obj = DataClass(MySelfSerializableType())
    assert obj.to_dict() == {"x": obj.x}
    assert DataClass.from_dict({"x": obj.x}) == obj


def test_annotated_generic_serializable_type_with_mixed_type_vars():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: MyAnnotatedGenericSerializableTypeWithMixedTypeVars[date]

    obj = DataClass.from_dict({"x": ["1", "2022-05-29", "2.3"]})
    # assert obj.x.value == (1, date(2022, 5, 29), 2.3)
    assert obj.to_dict() == {"x": [1, "2022-05-29", 2.3]}


def test_alias():
    x1 = Alias("x")
    x2 = Alias("x")
    y = Alias("y")

    assert x1 == x2
    assert hash(x1) == hash(x2)

    assert x1 != y
    assert x1 != "x"
    assert hash(x1) != hash(y)

    assert str(y) == "Alias(name='y')"
    assert repr(y) == "Alias(name='y')"



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Optional, Union

import pyperf
from marshmallow import Schema, fields, post_load

from benchmark.common import AbstractBenchmark


class IssueState(Enum):
    OPEN = "open"
    CLOSED = "closed"


class MilestoneState(Enum):
    OPEN = "open"
    CLOSED = "closed"


class IssueStateReason(Enum):
    COMPLETED = "completed"
    REOPENED = "reopened"
    NOT_PLANNED = "not_planned"


class AuthorAssociation(Enum):
    COLLABORATOR = "COLLABORATOR"
    CONTRIBUTOR = "CONTRIBUTOR"
    FIRST_TIMER = "FIRST_TIMER"
    FIRST_TIME_CONTRIBUTOR = "FIRST_TIME_CONTRIBUTOR"
    MANNEQUIN = "MANNEQUIN"
    MEMBER = "MEMBER"
    NONE = "NONE"
    OWNER = "OWNER"


@dataclass
class User:
    login: str
    id: int
    node_id: str
    avatar_url: str
    gravatar_id: Optional[str]
    url: str
    html_url: str
    followers_url: str
    following_url: str
    gists_url: str
    starred_url: str
    subscriptions_url: str
    organizations_url: str
    repos_url: str
    events_url: str
    received_events_url: str
    type: str
    site_admin: bool
    name: Optional[str] = None
    email: Optional[str] = None
    starred_at: Optional[datetime] = None


@dataclass
class IssueLabel:
    id: int
    node_id: str
    url: str
    name: str
    description: Optional[str]
    color: Optional[str]
    default: bool


@dataclass
class Milestone:
    url: str
    html_url: str
    labels_url: str
    id: int
    node_id: str
    number: int
    title: str
    description: Optional[str]
    creator: Optional[User]
    open_issues: int
    closed_issues: int
    created_at: datetime
    updated_at: datetime
    closed_at: Optional[datetime]
    due_on: Optional[datetime]
    state: MilestoneState = MilestoneState.OPEN


@dataclass
class Reactions:
    url: str
    total_count: int
    plus_one: int
    minus_one: int
    laugh: int
    confused: int
    heart: int
    hooray: int
    eyes: int
    rocket: int


@dataclass
class Issue:
    id: int
    node_id: str
    url: str
    repository_url: str
    labels_url: str
    comments_url: str
    events_url: str
    html_url: str
    number: int
    state: IssueState
    state_reason: Optional[IssueStateReason]
    title: str
    body: Optional[str]
    user: Optional[User]
    labels: list[Union[IssueLabel, str]]
    assignee: Optional[User]
    assignees: Optional[list[User]]
    milestone: Optional[Milestone]
    locked: bool
    active_lock_reason: Optional[str]
    comments: int
    closed_at: Optional[datetime]
    created_at: datetime
    updated_at: datetime
    closed_by: Optional[User]
    author_association: AuthorAssociation
    draft: bool = False
    body_html: Optional[str] = None
    body_text: Optional[str] = None
    timeline_url: Optional[str] = None
    reactions: Optional[Reactions] = None


class UserSchema(Schema):
    login = fields.Str()
    id = fields.Int()
    node_id = fields.Str()
    avatar_url = fields.Str()
    gravatar_id = fields.Str(allow_none=True)
    url = fields.Str()
    html_url = fields.Str()
    followers_url = fields.Str()
    following_url = fields.Str()
    gists_url = fields.Str()
    starred_url = fields.Str()
    subscriptions_url = fields.Str()
    organizations_url = fields.Str()
    repos_url = fields.Str()
    events_url = fields.Str()
    received_events_url = fields.Str()
    type = fields.Str()
    site_admin = fields.Bool()
    name = fields.Str(allow_none=True, load_default=True)
    email = fields.Str(allow_none=True, load_default=True)
    starred_at = fields.DateTime(allow_none=True, load_default=None)

    @post_load
    def _make_model(self, data, **kwargs):
        return User(**data)


class IssueLabelSchema(Schema):
    id = fields.Int()
    node_id = fields.Str()
    url = fields.Str()
    name = fields.Str()
    description = fields.Str(allow_none=True)
    color = fields.Str(allow_none=True)
    default = fields.Bool()

    @post_load
    def _make_model(self, data, **kwargs):
        return IssueLabel(**data)


class MilestoneSchema(Schema):
    url = fields.Str()
    html_url = fields.Str()
    labels_url = fields.Str()
    id = fields.Int()
    node_id = fields.Str()
    number = fields.Int()
    title = fields.Str()
    description = fields.Str(allow_none=True)
    creator = fields.Nested(UserSchema(), allow_none=True)
    open_issues = fields.Int()
    closed_issues = fields.Int()
    created_at = fields.DateTime()
    updated_at = fields.DateTime()
    closed_at = fields.DateTime(allow_none=True)
    due_on = fields.DateTime(allow_none=True)
    state = fields.Enum(
        MilestoneState, load_default=MilestoneState.OPEN, by_value=True
    )

    @post_load
    def _make_model(self, data, **kwargs):
        return Milestone(**data)


class ReactionsSchema(Schema):
    url = fields.Str()
    total_count = fields.Int()
    plus_one = fields.Int(data_key="+1")
    minus_one = fields.Int(data_key="-1")
    laugh = fields.Int()
    confused = fields.Int()
    heart = fields.Int()
    hooray = fields.Int()
    eyes = fields.Int()
    rocket = fields.Int()

    @post_load
    def _make_model(self, data, **kwargs):
        return Reactions(**data)


class IssueLabelSchemaOrStr(fields.Method):
    def __init__(self):
        self._issue_label_schema = IssueLabelSchema()
        super().__init__()

    def _deserialize(self, value, attr, data, **kwargs):
        try:
            return self._issue_label_schema.load(value)
        except ValueError:
            return str(value)

    def _serialize(self, value, attr, obj, **kwargs):
        return (
            self._issue_label_schema.dump(value)
            if isinstance(value, IssueLabel)
            else value
        )


class IssueSchema(Schema):
    id = fields.Int()
    node_id = fields.Str()
    url = fields.Str()
    repository_url = fields.Str()
    labels_url = fields.Str()
    comments_url = fields.Str()
    events_url = fields.Str()
    html_url = fields.Str()
    number = fields.Int()
    state = fields.Enum(IssueState, by_value=True)
    state_reason = fields.Enum(
        IssueStateReason, allow_none=True, by_value=True
    )
    title = fields.Str()
    body = fields.Str(allow_none=True)
    user = fields.Nested(UserSchema(), allow_none=True)
    labels = fields.List(IssueLabelSchemaOrStr())
    assignee = fields.Nested(UserSchema(), allow_none=True)
    assignees = fields.List(fields.Nested(UserSchema()))
    milestone = fields.Nested(MilestoneSchema(), allow_none=True)
    locked = fields.Bool()
    active_lock_reason = fields.Str(allow_none=True)
    comments = fields.Int()
    closed_at = fields.DateTime(allow_none=True)
    created_at = fields.DateTime()
    updated_at = fields.DateTime()
    closed_by = fields.Nested(UserSchema(), allow_none=True)
    author_association = fields.Enum(AuthorAssociation, by_value=True)
    draft = fields.Bool(load_default=False)
    body_html = fields.Str(allow_none=True, load_default=None)
    body_text = fields.Str(allow_none=True, load_default=None)
    timeline_url = fields.Str(allow_none=True, load_default=None)
    reactions = fields.Nested(
        ReactionsSchema(), allow_none=True, load_default=None
    )

    @post_load
    def _make_model(self, data, **kwargs):
        return Issue(**data)


class Benchmark(AbstractBenchmark):
    LIBRARY = "marshmallow"

    def __init__(self, runner: pyperf.Runner):
        super().__init__(runner)
        self._issue_schema = IssueSchema()

    def warmup(self, data) -> None:
        self._issue_schema.dump(self._issue_schema.load(data))

    def run_loader(self, data) -> pyperf.Benchmark:
        return self._bench_loader_func(self._issue_schema.load, data)

    def run_dumper(self, data) -> pyperf.Benchmark:
        obj = self._issue_schema.load(data)
        return self._bench_dumper_func(self._issue_schema.dump, obj)



from dataclasses import dataclass, field
from email.policy import default
from importlib.metadata import metadata
from typing import Optional

import pytest
from typing_extensions import Annotated

from mashumaro import DataClassDictMixin, pass_through
from mashumaro.config import (
    TO_DICT_ADD_BY_ALIAS_FLAG,
    TO_DICT_ADD_OMIT_NONE_FLAG,
    BaseConfig,
)
from mashumaro.exceptions import MissingField
from mashumaro.types import Alias


@dataclass
class Aliased(DataClassDictMixin):
    a: int = field(metadata={"alias": "alias_a"})
    b: Annotated[int, Alias("alias_b")]


@dataclass
class AliasedWithSerializeByAliasFlag(DataClassDictMixin):
    a: int = field(metadata={"alias": "alias_a"})
    b: Annotated[int, Alias("alias_b")]

    class Config(BaseConfig):
        code_generation_options = [TO_DICT_ADD_BY_ALIAS_FLAG]


def test_alias():
    assert Aliased.from_dict({"alias_a": 123, "alias_b": 456}) == Aliased(
        a=123, b=456
    )
    with pytest.raises(MissingField):
        assert Aliased.from_dict({"a": 123, "alias_b": 456})
    with pytest.raises(MissingField):
        assert Aliased.from_dict({"alias_a": 123, "b": 456})


def test_alias_with_default():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: int = field(default=111, metadata={"alias": "alias_a"})
        b: Annotated[int, Alias("alias_b")] = 222

    assert DataClass.from_dict({"alias_a": 123, "alias_b": 456}) == DataClass(
        a=123, b=456
    )
    assert DataClass.from_dict({}) == DataClass(a=111, b=222)


def test_alias_with_omit_none():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: Optional[int] = field(default=None, metadata={"alias": "alias_a"})
        b: Annotated[Optional[int], Alias("alias_b")] = None

        class Config(BaseConfig):
            code_generation_options = [
                TO_DICT_ADD_BY_ALIAS_FLAG,
                TO_DICT_ADD_OMIT_NONE_FLAG,
            ]

    instance = DataClass()
    assert instance.to_dict(omit_none=True) == {}
    assert instance.to_dict(by_alias=True) == {
        "alias_a": None,
        "alias_b": None,
    }
    assert instance.to_dict(omit_none=True, by_alias=True) == {}
    instance = DataClass(a=123, b=456)
    assert instance.to_dict(omit_none=True) == {"a": 123, "b": 456}
    assert instance.to_dict(by_alias=True) == {"alias_a": 123, "alias_b": 456}
    assert instance.to_dict(omit_none=True, by_alias=True) == {
        "alias_a": 123,
        "alias_b": 456,
    }


def test_serialize_by_alias_config_option():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: int = field(metadata={"alias": "alias_a"})
        b: Annotated[int, Alias("alias_b")]

        class Config(BaseConfig):
            serialize_by_alias = True

    assert DataClass(123, 456).to_dict() == {"alias_a": 123, "alias_b": 456}


def test_serialize_by_alias_code_generation_flag():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int = field(metadata={"alias": "alias_x"})
        y: Annotated[int, Alias("alias_y")]

        class Config(BaseConfig):
            code_generation_options = [TO_DICT_ADD_BY_ALIAS_FLAG]

    instance = DataClass(x=123, y=456)
    assert instance.to_dict() == {"x": 123, "y": 456}
    assert instance.to_dict(by_alias=True) == {"alias_x": 123, "alias_y": 456}


def test_serialize_by_alias_code_generation_flag_without_alias():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int

        class Config(BaseConfig):
            code_generation_options = [TO_DICT_ADD_BY_ALIAS_FLAG]

    instance = DataClass(x=123)
    assert instance.to_dict() == {"x": 123}
    assert instance.to_dict(by_alias=True) == {"x": 123}


def test_no_serialize_by_alias_code_generation_flag():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: int = field(metadata={"alias": "alias_x"})
        y: Annotated[int, Alias("alias_y")]

    instance = DataClass(x=123, y=456)
    assert instance.to_dict() == {"x": 123, "y": 456}
    with pytest.raises(TypeError):
        instance.to_dict(by_alias=True)


def test_serialize_by_alias_flag_for_inner_class_without_it():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Aliased = field(metadata={"alias": "alias_x"})
        y: Annotated[Aliased, Alias("alias_y")]

        class Config(BaseConfig):
            code_generation_options = [TO_DICT_ADD_BY_ALIAS_FLAG]

    instance = DataClass(Aliased(a=1, b=2), Aliased(a=3, b=4))
    assert instance.to_dict() == {"x": {"a": 1, "b": 2}, "y": {"a": 3, "b": 4}}
    assert instance.to_dict(by_alias=True) == {
        "alias_x": {"a": 1, "b": 2},
        "alias_y": {"a": 3, "b": 4},
    }


def test_serialize_by_alias_flag_for_inner_class_with_it():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: AliasedWithSerializeByAliasFlag = field(
            metadata={"alias": "alias_x"}
        )
        y: Annotated[AliasedWithSerializeByAliasFlag, Alias("alias_y")]

        class Config(BaseConfig):
            code_generation_options = [TO_DICT_ADD_BY_ALIAS_FLAG]

    instance = DataClass(
        AliasedWithSerializeByAliasFlag(a=1, b=2),
        AliasedWithSerializeByAliasFlag(a=3, b=4),
    )
    assert instance.to_dict() == {"x": {"a": 1, "b": 2}, "y": {"a": 3, "b": 4}}
    assert instance.to_dict(by_alias=True) == {
        "alias_x": {"alias_a": 1, "alias_b": 2},
        "alias_y": {"alias_a": 3, "alias_b": 4},
    }


def test_aliases_in_config():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: int = 111
        b: Annotated[int, Alias("alias_b")] = 222

        class Config(BaseConfig):
            aliases = {"a": "alias_a", "b": "alias_c"}
            serialize_by_alias = True

    assert DataClass.from_dict({"alias_a": 123, "alias_b": 456}) == DataClass(
        a=123, b=456
    )
    assert DataClass.from_dict({}) == DataClass(a=111, b=222)
    assert DataClass(a=123, b=456).to_dict() == {
        "alias_a": 123,
        "alias_b": 456,
    }


def test_by_alias_with_serialize_by_alias():
    @dataclass
    class DataClass(DataClassDictMixin):
        a: int = field(metadata={"alias": "alias_a"})
        b: Annotated[int, Alias("alias_b")]

        class Config(BaseConfig):
            serialize_by_alias = True
            code_generation_options = [TO_DICT_ADD_BY_ALIAS_FLAG]

    instance = DataClass(a=123, b=456)
    assert DataClass.from_dict({"alias_a": 123, "alias_b": 456}) == instance
    assert instance.to_dict() == {"alias_a": 123, "alias_b": 456}
    assert instance.to_dict(by_alias=False) == {"a": 123, "b": 456}


def test_no_serialize_by_alias_with_serialize_by_alias_and_optional():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Optional[int] = field(metadata={"alias": "alias_x"})
        y: Annotated[Optional[int], Alias("alias_y")]

        class Config(BaseConfig):
            serialize_by_alias = True

    assert DataClass(x=123, y=456).to_dict() == {
        "alias_x": 123,
        "alias_y": 456,
    }
    assert DataClass(x=None, y=None).to_dict() == {
        "alias_x": None,
        "alias_y": None,
    }


def test_by_field_with_allow_deserialization_not_by_alias():
    @dataclass
    class DataClass(DataClassDictMixin):
        a1: int = field(metadata={"alias": "alias_a1"})
        a2: Annotated[int, Alias("alias_a2")]
        b1: Optional[int] = field(metadata={"alias": "alias_b1"})
        b2: Annotated[Optional[int], Alias("alias_b2")]
        c1: Optional[str] = field(metadata={"alias": "alias_c1"})
        c2: Annotated[Optional[str], Alias("alias_c2")]
        d1: int = field(
            metadata={"alias": "alias_d1", "deserialize": pass_through}
        )
        d2: Annotated[int, Alias("alias_d2")] = field(
            metadata={"deserialize": pass_through}
        )
        e1: int = field(metadata={"alias": "alias_e1"}, default=5)
        e2: Annotated[int, Alias("alias_e2")] = 5
        f1: Optional[int] = field(metadata={"alias": "alias_f1"}, default=6)
        f2: Annotated[Optional[int], Alias("alias_f2")] = 6
        g1: Optional[str] = field(metadata={"alias": "alias_g1"}, default="7")
        g2: Annotated[Optional[str], Alias("alias_g2")] = "7"
        h1: int = field(
            metadata={"alias": "alias_h1", "deserialize": pass_through},
            default=8,
        )
        h2: Annotated[int, Alias("alias_h2")] = field(
            metadata={"deserialize": pass_through}, default=8
        )

        class Config(BaseConfig):
            serialize_by_alias = True
            code_generation_options = [TO_DICT_ADD_BY_ALIAS_FLAG]
            allow_deserialization_not_by_alias = True

    instance = DataClass(a1=1, a2=1, b1=2, b2=2, c1="3", c2="3", d1=4, d2=4)
    assert (
        DataClass.from_dict(
            {
                "a1": 1,
                "a2": 1,
                "b1": 2,
                "b2": 2,
                "c1": "3",
                "c2": "3",
                "d1": 4,
                "d2": 4,
                "e1": 5,
                "e2": 5,
                "f1": 6,
                "f2": 6,
                "g1": "7",
                "g2": "7",
                "h1": 8,
                "h2": 8,
            }
        )
        == instance
    )
    assert (
        DataClass.from_dict(
            {
                "alias_a1": 1,
                "alias_a2": 1,
                "alias_b1": 2,
                "alias_b2": 2,
                "alias_c1": "3",
                "alias_c2": "3",
                "alias_d1": 4,
                "alias_d2": 4,
                "alias_e1": 5,
                "alias_e2": 5,
                "alias_f1": 6,
                "alias_f2": 6,
                "alias_g1": "7",
                "alias_g2": "7",
                "alias_h1": 8,
                "alias_h2": 8,
            }
        )
        == instance
    )
    assert instance.to_dict() == {
        "alias_a1": 1,
        "alias_a2": 1,
        "alias_b1": 2,
        "alias_b2": 2,
        "alias_c1": "3",
        "alias_c2": "3",
        "alias_d1": 4,
        "alias_d2": 4,
        "alias_e1": 5,
        "alias_e2": 5,
        "alias_f1": 6,
        "alias_f2": 6,
        "alias_g1": "7",
        "alias_g2": "7",
        "alias_h1": 8,
        "alias_h2": 8,
    }
    assert instance.to_dict(by_alias=False) == {
        "a1": 1,
        "a2": 1,
        "b1": 2,
        "b2": 2,
        "c1": "3",
        "c2": "3",
        "d1": 4,
        "d2": 4,
        "e1": 5,
        "e2": 5,
        "f1": 6,
        "f2": 6,
        "g1": "7",
        "g2": "7",
        "h1": 8,
        "h2": 8,
    }


def test_order_of_metadata_and_annotated():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: Annotated[int, Alias("foo")] = field(metadata={"alias": "bar"})

        class Config(BaseConfig):
            serialize_by_alias = True

    instance = DataClass(42)
    assert DataClass.from_dict({"bar": 42}) == instance
    assert instance.to_dict() == {"bar": 42}



from dataclasses import dataclass
from datetime import date, datetime
from typing import Any, Mapping

from mashumaro import DataClassDictMixin
from mashumaro.mixins.json import DataClassJSONMixin
from mashumaro.types import SerializableType
from tests.entities import MyGenericDataClass, SerializableTypeGenericList


@dataclass
class Foo[T](DataClassJSONMixin):
    x: T
    y: "Foo[Any] | None"


@dataclass
class Bar(Foo):
    pass


@dataclass
class GenericSerializableType[T](SerializableType):
    value: object

    def _serialize(self):
        return self.value

    @classmethod
    def _deserialize(cls, value):
        return cls(value)


def test_one_generic():
    @dataclass
    class A[T]:
        x: T

    @dataclass
    class B(A[datetime], DataClassDictMixin):
        pass

    obj = B(datetime(2021, 8, 15))
    assert B.from_dict({"x": "2021-08-15T00:00:00"}) == obj
    assert obj.to_dict() == {"x": "2021-08-15T00:00:00"}


def test_one_generic_list():
    @dataclass
    class A[T](list[T]):
        x: list[T]

    @dataclass
    class B(A[datetime], DataClassDictMixin):
        pass

    obj = B(x=[datetime(2021, 8, 15)])
    assert B.from_dict({"x": ["2021-08-15T00:00:00"]}) == obj
    assert obj.to_dict() == {"x": ["2021-08-15T00:00:00"]}


def test_two_generics():
    @dataclass
    class A1[T]:
        x: list[T]

    @dataclass
    class A2[T, S]:
        y: Mapping[T, S]

    @dataclass
    class B(A1[datetime], A2[datetime, date], DataClassDictMixin):
        pass

    obj = B(
        x=[datetime(2021, 8, 15), datetime(2021, 8, 16)],
        y={datetime(2021, 8, 17): date(2021, 8, 18)},
    )
    dump = {
        "x": ["2021-08-15T00:00:00", "2021-08-16T00:00:00"],
        "y": {"2021-08-17T00:00:00": "2021-08-18"},
    }
    assert B.from_dict(dump) == obj
    assert obj.to_dict() == dump


def test_partially_concrete_generic():
    @dataclass
    class A[T, S]:
        x: Mapping[T, S]

    @dataclass
    class B[S](A[datetime, S], DataClassDictMixin):
        pass

    obj = B(x={datetime(2022, 11, 21): "3.14"})
    assert B.from_dict({"x": {"2022-11-21T00:00:00": "3.14"}}) == obj
    assert obj.to_dict() == {"x": {"2022-11-21T00:00:00": "3.14"}}


def test_partially_concrete_generic_with_bound():
    @dataclass
    class A[T, P: (Mapping[int, int], list[float])]:
        x: Mapping[T, P]

    @dataclass
    class B[P: (Mapping[int, int], list[float])](
        A[date, P], DataClassDictMixin
    ):
        pass

    obj1 = B(x={date(2022, 11, 21): {1: 2, 3: 4}})
    assert B.from_dict({"x": {"2022-11-21": {"1": "2", "3": "4"}}}) == obj1
    assert obj1.to_dict() == {"x": {"2022-11-21": {1: 2, 3: 4}}}
    obj2 = B(x={date(2022, 11, 21): [1.1, 3.3]})
    assert (
        B.from_dict({"x": {"2022-11-21": {"1.1": "2.2", "3.3": "4.4"}}})
        == obj2
    )
    assert obj2.to_dict() == {"x": {"2022-11-21": [1.1, 3.3]}}
    obj3 = B(x={date(2022, 11, 21): [1.1, 2.2, 3.3, 4.4]})
    assert (
        B.from_dict({"x": {"2022-11-21": ["1.1", "2.2", "3.3", "4.4"]}})
        == obj3
    )
    assert obj3.to_dict() == {"x": {"2022-11-21": [1.1, 2.2, 3.3, 4.4]}}


def test_concrete_generic_with_different_type_var():
    @dataclass
    class A[T]:
        x: T

    @dataclass
    class B[P: (Mapping[int, int], list[float])](A[P], DataClassDictMixin):
        pass

    obj = B.from_dict({"x": {"1": "2", "3": "4"}})
    assert obj == B(x={1: 2, 3: 4})
    obj = B.from_dict({"x": {"1.1": "2.2", "3.3": "4.4"}})
    assert obj == B(x=[1.1, 3.3])
    obj = B.from_dict({"x": ["1.1", "2.2", "3.3", "4.4"]})
    assert obj == B(x=[1.1, 2.2, 3.3, 4.4])


def test_loose_generic_info_with_any_type():
    @dataclass
    class A[T]:
        x: T

    @dataclass
    class B(A, DataClassDictMixin):
        pass

    obj = B.from_dict({"x": {"1.1": "2.2", "3.3": "4.4"}})
    assert obj == B(x={"1.1": "2.2", "3.3": "4.4"})
    obj = B.from_dict({"x": ["1.1", "2.2", "3.3", "4.4"]})
    assert obj == B(x=["1.1", "2.2", "3.3", "4.4"])


def test_loose_generic_info_with_bound():
    @dataclass
    class A[P: (Mapping[int, int], list[float])]:
        x: P

    @dataclass
    class B(A, DataClassDictMixin):
        pass

    obj = B.from_dict({"x": {"1": "2", "3": "4"}})
    assert obj == B(x={1: 2, 3: 4})
    obj = B.from_dict({"x": {"1.1": "2.2", "3.3": "4.4"}})
    assert obj == B(x=[1.1, 3.3])
    obj = B.from_dict({"x": ["1.1", "2.2", "3.3", "4.4"]})
    assert obj == B(x=[1.1, 2.2, 3.3, 4.4])


def test_loose_generic_info_in_first_generic():
    @dataclass
    class A[P: (Mapping[int, int], list[float])]:
        x: P

    @dataclass
    class B(A):
        pass

    @dataclass
    class C[P: (Mapping[int, int], list[float])](B):
        y: P

    @dataclass
    class D(C[list[float]], DataClassDictMixin):
        pass

    obj = D.from_dict({"x": {"1": "2"}, "y": {"3.3": "4.4"}})
    assert obj == D(x={1: 2}, y=[3.3])
    obj = D.from_dict({"x": {"1.1": "2.2"}, "y": {"3.3": "4.4"}})
    assert obj == D(x=[1.1], y=[3.3])


def test_not_dataclass_generic():
    class MyGeneric[P: (Mapping[int, int], list[float]), T]:
        pass

    @dataclass
    class GenericDataClass[P: (Mapping[int, int], list[float])]:
        x: P

    @dataclass
    class DataClass[P: (Mapping[int, int], list[float]), T](
        MyGeneric[P, T], GenericDataClass[P]
    ):
        pass

    @dataclass
    class ConcreteDataClass(DataClass[list[float], float], DataClassDictMixin):
        pass

    obj = ConcreteDataClass.from_dict({"x": {"1": "2", "3": "4"}})
    assert obj == ConcreteDataClass(x=[1.0, 3.0])


def test_generic_dataclass_as_field_type():
    @dataclass
    class DataClass(DataClassDictMixin):
        date: MyGenericDataClass[date]
        str: MyGenericDataClass[str]

    obj = DataClass(
        date=MyGenericDataClass(date(2021, 9, 14)),
        str=MyGenericDataClass("2021-09-14"),
    )
    dictionary = {"date": {"x": "2021-09-14"}, "str": {"x": "2021-09-14"}}
    assert DataClass.from_dict(dictionary) == obj
    assert obj.to_dict() == dictionary


def test_serializable_type_generic_class():
    @dataclass
    class DataClass(DataClassDictMixin):
        x: SerializableTypeGenericList[str]

    obj = DataClass(SerializableTypeGenericList(["a", "b", "c"]))
    assert DataClass.from_dict({"x": ["a", "b", "c"]}) == obj
    assert obj.to_dict() == {"x": ["a", "b", "c"]}


def test_self_referenced_generic_no_max_recursion_error():
    obj = Bar(42, Foo(33, None))
    assert obj.to_dict() == {"x": 42, "y": {"x": 33, "y": None}}
    assert Bar.from_dict({"x": 42, "y": {"x": 33, "y": None}}) == obj
    assert obj.to_json() == '{"x": 42, "y": {"x": 33, "y": null}}'
    assert Bar.from_json('{"x": 42, "y": {"x": 33, "y": null}}') == obj


# Regression test for https://github.com/Fatal1ty/mashumaro/issues/274.
def test_generic_serializable_type_getitem():
    @dataclass
    class DataClass(DataClassDictMixin):
        # Simply specializing the type would lead to an AttributeError.
        x: GenericSerializableType[int]



from dataclasses import dataclass
from typing import Optional

from mashumaro import DataClassDictMixin
from mashumaro.config import (
    TO_DICT_ADD_BY_ALIAS_FLAG,
    TO_DICT_ADD_OMIT_NONE_FLAG,
    BaseConfig,
)


@dataclass
class A(DataClassDictMixin):
    x: Optional[int] = None

    class Config(BaseConfig):
        aliases = {"x": "x_alias"}
        code_generation_options = [
            TO_DICT_ADD_OMIT_NONE_FLAG,
            TO_DICT_ADD_BY_ALIAS_FLAG,
        ]


@dataclass
class B(DataClassDictMixin):
    a: Optional[A] = None

    class Config(BaseConfig):
        aliases = {"a": "a_alias"}
        code_generation_options = [
            TO_DICT_ADD_OMIT_NONE_FLAG,
            TO_DICT_ADD_BY_ALIAS_FLAG,
        ]


def test_passing_flags_if_parent_has_them():
    @dataclass
    class WithFlags(DataClassDictMixin):
        b: B

        class Config(BaseConfig):
            code_generation_options = [
                TO_DICT_ADD_OMIT_NONE_FLAG,
                TO_DICT_ADD_BY_ALIAS_FLAG,
            ]

    assert WithFlags.from_dict({"b": {"a": {"x": None}}}) == WithFlags(
        b=B(a=None)
    )
    obj = WithFlags.from_dict({"b": {"a_alias": {"x": None}}})
    assert obj == WithFlags(b=B(a=A(x=None)))
    assert obj.to_dict() == {"b": {"a": {"x": None}}}
    assert obj.to_dict(by_alias=True) == {"b": {"a_alias": {"x_alias": None}}}
    assert obj.to_dict(by_alias=True, omit_none=True) == {"b": {"a_alias": {}}}


def test_passing_flags_if_parent_does_not_have_them():
    @dataclass
    class WithoutFlags(DataClassDictMixin):
        b: B

    assert WithoutFlags.from_dict({"b": {"a": {"x": None}}}) == WithoutFlags(
        b=B(a=None)
    )
    obj = WithoutFlags.from_dict({"b": {"a_alias": {"x": None}}})
    assert obj == WithoutFlags(b=B(a=A(x=None)))
    assert obj.to_dict() == {"b": {"a": {"x": None}}}



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import TypedDict

from mashumaro import DataClassDictMixin


class Bar(TypedDict):
    baz: str


class Foo(TypedDict):
    bar: Bar


@dataclass
class FooBar(DataClassDictMixin):
    foo: Foo


def test_foobar():
    obj = FooBar.from_dict({"foo": {"bar": {"baz": "baz"}}})
    assert obj.foo["bar"]["baz"] == "baz"
    assert obj.to_dict() == {"foo": {"bar": {"baz": "baz"}}}



from dataclasses import dataclass

from mashumaro import DataClassDictMixin
from mashumaro.codecs import BasicDecoder, BasicEncoder

type JSON = str | int | float | bool | dict[str, JSON] | list[JSON] | None


@dataclass
class MyClass:
    x: str
    y: JSON


def test_encoder_with_recursive_union():
    encoder = BasicEncoder(JSON)
    assert encoder.encode(
        {"x": [{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}]}
    ) == {"x": [{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}]}


def test_encoder_with_recursive_union_in_dataclass():
    encoder = BasicEncoder(MyClass)
    assert encoder.encode(
        MyClass(
            x="x", y={"x": [{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}]}
        )
    ) == {
        "x": "x",
        "y": {"x": [{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}]},
    }


def test_decoder_with_recursive_union():
    decoder = BasicDecoder(JSON)
    assert decoder.decode(
        {"x": [{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}]}
    ) == {"x": [{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}]}


def test_decoder_with_recursive_union_in_dataclass():
    decoder = BasicDecoder(MyClass)
    assert decoder.decode(
        {
            "x": "x",
            "y": {"x": [{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}]},
        }
    ) == MyClass(
        x="x", y={"x": [{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}]}
    )


def test_dataclass_dict_mixin_with_recursive_union():
    @dataclass
    class MyClassWithMixin(DataClassDictMixin):
        x: str
        y: JSON

    assert MyClassWithMixin(
        x="x", y={"x": [{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}]}
    ).to_dict() == {
        "x": "x",
        "y": {"x": [{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}]},
    }
    assert MyClassWithMixin.from_dict(
        {
            "x": "x",
            "y": {"x": [{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}]},
        }
    ) == MyClassWithMixin(
        x="x", y={"x": [{"x": {"x": [{"x": ["x", 1, 1.0, True, None]}]}}]}
    )



from datetime import date
from typing import List

from mashumaro.codecs.orjson import (
    ORJSONDecoder,
    ORJSONEncoder,
    json_decode,
    json_encode,
)
from mashumaro.dialect import Dialect


class MyDialect(Dialect):
    serialization_strategy = {
        date: {
            "serialize": date.toordinal,
            "deserialize": date.fromordinal,
        },
    }


def test_json_decode():
    assert json_decode('["2023-09-22", "2023-09-23"]', List[date]) == [
        date(2023, 9, 22),
        date(2023, 9, 23),
    ]


def test_json_encode():
    assert (
        json_encode([date(2023, 9, 22), date(2023, 9, 23)], List[date])
        == b'["2023-09-22","2023-09-23"]'
    )


def test_decoder_with_default_dialect():
    decoder = ORJSONDecoder(List[date], default_dialect=MyDialect)
    assert decoder.decode("[738785, 738786]") == [
        date(2023, 9, 22),
        date(2023, 9, 23),
    ]


def test_encoder_with_default_dialect():
    encoder = ORJSONEncoder(List[date], default_dialect=MyDialect)
    assert (
        encoder.encode([date(2023, 9, 22), date(2023, 9, 23)])
        == b"[738785,738786]"
    )



import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help="input file name")
    parser.add_argument("output", help="output file name")
    args = parser.parse_args()

    if not args.input or not args.output:
        print("Please provide both input and output file names.")
        exit(1)

    # Replace all occurrences of "#333333" with "#C9D1D9" for dark theme
    with open(args.input) as infile, open(args.output, "w") as outfile:
        for line in infile:
            outfile.write(line.replace("#333333", "#C9D1D9"))
